/******************************************************************************
 *
 * Copyright (c) 1995-2002 Palm, Inc. or its subsidiaries.
 * All rights reserved.
 *
 * File: DateDB.h
 *
 * Release: from Palm OS SDK 5.0 (111823)
 *
 * Description:
 *		Header for the Appointment Manager
 *
 * History:
 *   	01/24/04  smb - Created
 *
 *****************************************************************************/

#ifndef __DATEDB_H__
#define __DATEDB_H__


/************************************************************
 *
 * Appointment Database constants.
 *
 *************************************************************/
#define apptMaxPerDay 		100		// max appointments displayable on a day.
#define apptNoTime			-1			// start time of an untimed appt.
#define apptNoEndDate		0xffff	// end date of appts that repeat forever
#define apptNoAlarm			-1
#define apptMaxEndTime		0x1737	// 11:55 pm, hours in high byte, minutes in low byte
#define apptDawnOfTime		0
#define apptEndOfTime		0xffffffff

#define apptMaxDisplayableAlarms 10


/************************************************************
 *
 * Appointment Database structures.
 *
 *************************************************************/
typedef struct {
	TimeType			startTime;			// Time the appointment starts
	TimeType			endTime;				// Time the appointment ends
	DateType			date;					// date of appointment
} ApptDateTimeType;

typedef struct {
	unsigned when			:1;			// set if when info changed (ApptChangeRecord)
	unsigned alarm			:1;			// set if record contains alarm info
	unsigned repeat			:1;			// set if record contains repeat info
	unsigned note			:1;			// set if record contains a note
	unsigned exceptions		:1;			// set if record contains exceptions list
	unsigned description	:1;			
} ApptDBRecordFlags;

typedef struct {
	ApptDateTimeType 	when;
	ApptDBRecordFlags	flags;	// A flag set for each  datum present
	char				firstField;
	UInt8				reserved;
} ApptPackedDBRecordType;

typedef ApptPackedDBRecordType * ApptPackedDBRecordPtr;

typedef enum alarmTypes {aauMinutes, aauHours, aauDays} AlarmUnitType;

enum repeatTypes {
	repeatNone, 
	repeatDaily, 
	repeatWeekly, 
	repeatMonthlyByDay, 
	repeatMonthlyByDate,
	repeatYearly
};
typedef enum repeatTypes RepeatType;

typedef struct {
	Int8				advance;			// Alarm advance (-1 = no alarm)
	AlarmUnitType		advanceUnit;	// minutes, hours, days
} AlarmInfoType;

typedef struct {
	RepeatType			repeatType;			// daily, weekly, monthlyByDay, etc.
	UInt8				reserved1;
	DateType			repeatEndDate;		// minus one if forever
	UInt8				repeatFrequency;	// i.e. every 2 days if repeatType daily
	UInt8				repeatOn;			// monthlyByDay and repeatWeekly only	
	UInt8 				repeatStartOfWeek;// repeatWeekly only
	UInt8				reserved2;
} RepeatInfoType;

typedef struct {
	UInt16    			numExceptions;
	DateType			exception;
} ExceptionsListType;

typedef struct {
	ApptDateTimeType	*when;
	AlarmInfoType		*alarm;
	RepeatInfoType		*repeat;
	ExceptionsListType	*exceptions;
	Char				*description;
	Char				*note;
} ApptDBRecordType;

typedef ApptDBRecordType * ApptDBRecordPtr;


/***********************************************************************
 *
 * Function Prototypes
 *
 ***********************************************************************/
Err CreateNewAppt (DmOpenRef dbP, DateType date,
 TimeType startTime, TimeType endTime, Char *description, UInt16 *recordNum);


#endif
