/******************************************************************************
 *
 * Copyright (c) 2004 PalmSpace/Sembe.
 *				All rights reserved.
 *
 * File: PsLink.cpp
 *
 * Description:
 *   This is the PsLink application's main module.
 *
 * History:
 *		Dec 29, 2003	Created by sembe
 *
 *****************************************************************************/

#include "PsLink.h"
#include "PsLinkLib.h"
#include "link.h"
#include "PsLinkPlugin.h"
#include "Resource.h"


/***********************************************************************
 *
 *	Internal Constants
 *
 ***********************************************************************/
#define ListID(c)	(((c) == 0)? MainMemoList :	\
					 ((c) == 1)? MainToDoList :	\
					 ((c) == 2)? MainAddressList : \
					 ((c) == 3)? MainDateList : \
					 ((c) == 4)? MainWebList : \
					 MainMailList)

#define CtrlID(c)	(((c) == 0)? MainMemoPopTrigger :	\
					 ((c) == 1)? MainToDoPopTrigger :	\
					 ((c) == 2)? MainAddressPopTrigger : \
					 ((c) == 3)? MainDatePopTrigger : \
					 ((c) == 4)? MainWebPopTrigger : \
					 MainMailPopTrigger)

#define Ctl2List(c)	(((c) == MainMemoPopTrigger)? MainMemoList :	\
					 ((c) == MainToDoPopTrigger)? MainToDoList :	\
					 ((c) == MainAddressPopTrigger)? MainAddressList : \
					 ((c) == MainDatePopTrigger)? MainDateList : \
					 ((c) == MainWebPopTrigger)? MainWebList : \
					 MainMailList)


/***********************************************************************
 *
 *	Internal Structures
 *
 ***********************************************************************/
typedef struct
{
	Char	name[dmDBNameLength];
	UInt32	creator;
} DBListType;

typedef struct
{
	DBListType		*list;
	Char			**names;
	Int16			num;
} DBListItemType;


/***********************************************************************
 *
 *	Global variables
 *
 ***********************************************************************/
static PsLinkPreferenceType	g_Prefs;

static DBListItemType		g_App;
static DBListItemType		g_DB;
static DBListItemType		g_Pnl;
static DBListItemType		g_Pin;

static UInt16				g_AppNum;
static Boolean				g_PnlListAvailable = false;
static TsmFepModeType		g_FepMode;


/***********************************************************************
 *
 * FUNCTION:    EnableNavigation
 *
 * DESCRIPTION: This routine sets the navigation state of the form to the object focus mode,
 *              if the device support 5-way navigation.
 *
 * PARAMETERS:  frmP - pointer to the form type struct
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		06/11/04	Initial Revision
 *
 ***********************************************************************/
static void EnableNavigation (FormPtr frmP)
{
	UInt32 val;

	if (FtrGet (hsFtrCreator, hsFtrIDNavigationSupported, &val) == errNone)
	{
		FrmSetNavState (frmP, kFrmNavStateFlagsObjectFocusMode);
	}
}


/***********************************************************************
 *
 * FUNCTION:    SetTriggerLabel
 *
 * DESCRIPTION: This routine sets the label for a trigger
 *				and sets the selection for a list.
 *
 * PARAMETERS:	ctrlID - ID of control type object
 *				listID - ID of list type object
 *				select - selected number of the list
 *				maxWidth - max width of label
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		01/15/04	Initial Revision
 *			smb		06/11/04	Modify to use CtlGetLabel().
 *
 ***********************************************************************/
static void SetTriggerLabel (UInt16 ctrlID, UInt16 listID, UInt16 select, Int16 maxWidth)
{
	ControlPtr	ctrlP;
	ListPtr		listP;
	Char		*label, *listStr;
	UInt16		labelLen;


	ctrlP = (ControlPtr)GetObjectPtr (ctrlID);
	listP = (ListPtr)GetObjectPtr (listID);
	listStr = LstGetSelectionText (listP, select);

	label = (Char *)CtlGetLabel (ctrlP);
	labelLen = StrLen (listStr);
	if (maxWidth != NULL
	 && FntWidthToOffset (listStr, labelLen, maxWidth, NULL, NULL) < labelLen)
	{
		labelLen = FntWidthToOffset (listStr, labelLen, maxWidth - FntCharWidth (chrEllipsis), NULL, NULL);
		StrNCopy (label, listStr, labelLen);
		label[labelLen] = chrEllipsis;
		label[labelLen + 1] = chrNull;
	}
	else
		StrCopy (label, listStr);

	LstSetSelection (listP, select);
	LstSetTopItem (listP, select);
	CtlSetLabel (ctrlP, label);
}


/***********************************************************************
 *
 * FUNCTION:    SetListItems
 *
 * DESCRIPTION: This routine sets list items.
 *
 * PARAMETERS:	listID - ID of list type object
 *				array - array for the list items
 *				numItems - number of items
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/10/04	Initial Revision
 *
 ***********************************************************************/
static void SetListItems (UInt16 listID, Char **array, Int16 numItems)
{
	ListPtr		listP;

	listP = (ListPtr)GetObjectPtr (listID);
	LstSetListChoices (listP, array, numItems);
	LstSetHeight (listP, numItems);
}


/***********************************************************************
 *
 * FUNCTION:    SelectList
 *
 * DESCRIPTION: This routine handles to popup and select list.
 *
 * PARAMETERS:	ctrlID - ID of control type object
 *				listID - ID of list type object
 *				maxWidth - max width of trigger
 *
 * RETURNED:	 nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/03	Initial Revision
 *
 ***********************************************************************/
static void SelectList (UInt16 ctrlID, UInt16 listID, UInt16 maxWidth)
{
	ListPtr		listP = (ListPtr)GetObjectPtr (listID);

	LstPopupList (listP);
	SetTriggerLabel (ctrlID, listID, LstGetSelection (listP), maxWidth);

}


/***********************************************************************
 *
 * FUNCTION:    SelectPanel
 *
 * DESCRIPTION: This routine handles to select preferences panel.
 *
 * PARAMETERS:	nothing
 *
 * RETURNED:	handled
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		01/25/04	Initial Revision
 *
 ***********************************************************************/
static Boolean SelectPanel (void)
{
	DmSearchStateType	state;
	UInt16				cardNo;
	LocalID				dbID;
	UInt32				creator;
	ListPtr				listP;
	Boolean				select;


	listP = (ListPtr)GetObjectPtr (MainPanelList);

	LstPopupList (listP);
	select = LstGetSelection (listP);

	creator = g_Pnl.list[select].creator;
	if (creator != appFileCreator)
	{
		DmGetNextDatabaseByTypeCreator (true, &state,
			 sysFileTPanel, creator, true, &cardNo, &dbID);
		SysUIAppSwitch (cardNo, dbID, sysAppLaunchCmdNormalLaunch, NULL);
	}

	return (true);
}


/***********************************************************************
 *
 * FUNCTION:    SetFieldString
 *
 * DESCRIPTION: This routine sets string to the field
 *
 * PARAMETERS:	objectID - id of the field object
 *				str - string to be set
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/29/03	Initial Revision
 *
 ***********************************************************************/
static void SetFieldString (UInt16 objectID, Char *str)
{
	FieldPtr	fldP;

	fldP = (FieldPtr)GetObjectPtr (objectID);
	FldDelete (fldP, 0, FldGetTextLength (fldP));
	FldInsert (fldP, str, StrLen (str));

}


/***********************************************************************
 *
 * FUNCTION:    GetFieldString
 *
 * DESCRIPTION: This routine gets the string from the specified field.
 *
 * PARAMETERS:	objectID - id of the field object
 *				maxLen - max length of the string
 *				string - pointer to the string
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		01/11/03	Initial Revision
 *
 ***********************************************************************/
static void GetFieldString (UInt16 objectID, UInt16 maxLen, Char *string)
{
	FieldPtr	fldP;
	UInt16		len;


	fldP = (FieldPtr)GetObjectPtr (objectID);
	len = FldGetTextLength (fldP);
	if (len)
	{
		len = min (len, maxLen);
		StrNCopy (string, FldGetTextPtr (fldP), len);
	}
	string[len] = chrNull;
}


/***********************************************************************
 *
 * FUNCTION:    RomVersionCompatible
 *
 * DESCRIPTION: This routine checks that a ROM version is meet your
 *              minimum requirement.
 *
 * PARAMETERS:  requiredVersion - minimum rom version required
 *                                (see sysFtrNumROMVersion in SystemMgr.h 
 *                                for format)
 *
 * RETURNED:    error code or zero if rom is compatible
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/29/03	Initial Revision
 *
 ***********************************************************************/
static Err RomVersionCompatible (UInt32 requiredVersion)
{
	UInt32 romVersion;

	// See if we're on in minimum required version of the ROM or later.
	FtrGet (sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	if (romVersion < requiredVersion)
	{
		FrmAlert (RomIncompatibleAlert);
		return sysErrRomIncompatible;
	}

	return errNone;
}


/***********************************************************************
 *
 * FUNCTION:    NotificationAvailable
 *
 * DESCRIPTION:	This routine checks whether notification is available or not.
 *
 * PARAMETERS:	nothing
 *
 * RETURNED:    error code or zero if available
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static Err NotificationAvailable (void)
{
	UInt32	value;

	return (FtrGet (sysFtrCreator,sysFtrNumNotifyMgrVersion, &value));
}


/***********************************************************************
 *
 * FUNCTION:    Enable
 *
 * DESCRIPTION: This routine register notify.
 *
 * PARAMETERS:	prefs : PsLinkPreferenceType structure
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/29/03	Initial Revision
 *			smb		02/28/05	Modify to use a parameter "prefs"
 *
 ***********************************************************************/
static void Enable (PsLinkPreferenceType *prefs)
{
	UInt16					cardNo;
	LocalID					dbID;
	KeyDownEventDataType	*keyAssignP;


	SysCurAppDatabase (&cardNo, &dbID);

	SysNotifyRegister (cardNo, dbID, sysNotifyEventDequeuedEvent,
			 NULL, sysNotifyNormalPriority, NULL);

	FtrSet (appFileCreator, FeatureState, IsON);
	FtrSet (appFileCreator, FeatureSelect, IsOFF);

	FtrPtrNew (appFileCreator, FeatureBackKey, sizeof (KeyDownEventDataType), (void **)&keyAssignP);
	DmWrite (keyAssignP, 0, &prefs->backKey, sizeof (KeyDownEventDataType));

	FtrPtrNew (appFileCreator, FeatureForwardKey, sizeof (KeyDownEventDataType), (void **)&keyAssignP);
	DmWrite (keyAssignP, 0, &prefs->forwardKey, sizeof (KeyDownEventDataType));
}


/***********************************************************************
 *
 * FUNCTION:    Disable
 *
 * DESCRIPTION: This routine unregister notify.
 *
 * PARAMETERS:	nothing
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/29/03	Initial Revision
 *
 ***********************************************************************/
static void Disable (void)
{
	UInt16		cardNo;
	LocalID		dbID;


	SysCurAppDatabase (&cardNo, &dbID);

	if (SysNotifyUnregister (cardNo, dbID, sysNotifyEventDequeuedEvent, sysNotifyNormalPriority)
	 & sysNotifyErrEntryNotFound)
		return;

	SysNotifyUnregister (cardNo, dbID, appFileCreator, sysNotifyNormalPriority);

	FtrUnregister (appFileCreator, FeatureState);
	FtrUnregister (appFileCreator, FeatureSelect);
	FtrUnregister (appFileCreator, FeatureSelStart);
	FtrUnregister (appFileCreator, FeatureSelEnd);
	FtrUnregister (appFileCreator, FeatureNavPressed);
	FtrPtrFree (appFileCreator, FeatureBackKey);
	FtrPtrFree (appFileCreator, FeatureForwardKey);
}


/***********************************************************************
 *
 * FUNCTION:	CompareF
 *
 * DESCRIPTION: This routine compares two strings
 *
 * PARAMETERS:  p1, p2 - pointer to a string to be compared
 *
 * RETURNED:    0, if match
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		01/18/04	Initial Revision
 *
 ***********************************************************************/
static Int16 CompareF (void *p1, void *p2, Int32 other)
{
    return StrCaselessCompare((Char *)p1, (Char *)p2);
}


/***********************************************************************
 *
 * FUNCTION:    GetDBList
 *
 * DESCRIPTION: This routine makes application list installed in device.
 *
 * PARAMETERS:  dbListP - pointer to DBListItemType struct.
 *				dbType - db type
 *				topItem - add this string at the top of list, if not NULL.
 *
 * RETURNED:    number of DBs
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		01/18/04	Initial Revision
 *			smb		02/10/04	Modify to set an array of db name
 *								 and deal with 'DATA' type db.
 *
 ***********************************************************************/
static UInt16 GetDBList (DBListItemType *dbListP, UInt16 dbType, Char *topItem)
{
	Err					err;
	DmSearchStateType	searchState;
	Boolean				newSearch;
	UInt16				numDBs;
	UInt16				cardNo;
	LocalID				dbID;
	Char				nameP[dmDBNameLength];
	UInt16				attr;
	UInt32				creator;
	UInt32				type;
	MemHandle			memH;
	Char				*memP;
	UInt16				i;


	if (*topItem != NULL)
	{
		memH = MemHandleNew (sizeof(DBListType) + 1);
		if (!memH)
			return 0;
		// Set "topItem" at the top of the list.
		memP = (Char *)MemHandleLock (memH);
		MemSet (memP, sizeof(DBListType), 0);
		StrCopy (memP, topItem);
		MemHandleUnlock (memH);
		numDBs = 1;
	}
	else
	{
		memH = MemHandleNew (1);
		if (!memH)
			return 0;
		numDBs = 0;
	}

	// Search all installed applications, and buffer these names and creator ids.
	newSearch = true;
	while (true)
	{
		err = DmGetNextDatabaseByTypeCreator (newSearch, &searchState,
									 NULL, NULL, false, &cardNo, &dbID);
		if (err)
			break;

		err = DmDatabaseInfo (cardNo, dbID, nameP, &attr,
			 NULL, NULL, NULL, NULL, NULL, NULL, NULL, &type, &creator);
		if (!err /*&& !(attr & dmHdrAttrHidden)*/
		 && (dbType == TYPE_APPL && type == sysFileTApplication
		 || dbType == TYPE_PANL && type == sysFileTPanel
		 || dbType == TYPE_DATA && IsData(type)
		 || dbType == TYPE_PLIN && type == PsLinkTPlugins))
		{
			if (MemHandleResize (memH, sizeof(DBListType) * (numDBs + 1) + 1) != errNone)
				break;
			if (attr & dmHdrAttrResDB)
			{// Try to get application name
				GetApplName (cardNo, dbID, nameP);
			}

			memP = (Char *)MemHandleLock (memH);
			StrCopy (memP + sizeof(DBListType) * numDBs, nameP);
			MemMove (memP + sizeof(DBListType) * (numDBs + 1) - sizeof(UInt32), (Char *)&creator, 4);
			MemHandleUnlock (memH);
			numDBs ++;
		}
		newSearch = false;
	}

	// Sort the list in alphabetical order.
	memP = (Char *)MemHandleLock (memH);
	if (*topItem != NULL)
		SysInsertionSort (memP + sizeof(DBListType), numDBs - 1, sizeof(DBListType), CompareF, NULL);
	else
		SysInsertionSort (memP, numDBs, sizeof(DBListType), CompareF, NULL);

	// Store the list items to a new chunk.
	dbListP->list = (DBListType *)MemPtrNew (sizeof(DBListType) * numDBs + 1);
	MemMove (dbListP->list, memP, sizeof(DBListType) * numDBs);

	MemHandleUnlock (memH);
	MemHandleFree (memH);

	dbListP->names = (Char **)MemPtrNew (sizeof(MemPtr) * numDBs);
	for (i = 0; i < numDBs; i ++)
		dbListP->names[i] = dbListP->list[i].name;

	dbListP->num = numDBs;

	return numDBs;
}


/***********************************************************************
 *
 * FUNCTION:    FreeDBList
 *
 * DESCRIPTION: This routine disposes a chunk for db list.
 *
 * PARAMETERS:  dbListP - pointer to DBListItemType struct.
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		06/06/04	Initial Revision
 *
 ***********************************************************************/
static void FreeDBList (DBListItemType dbList)
{
	if (dbList.list)
		MemPtrFree (dbList.list);
	if (dbList.names)
		MemPtrFree (dbList.names);
}


/***********************************************************************
 *
 * FUNCTION:    FindListItemByCreator
 *
 * DESCRIPTION: This routine finds item number in list by creator.
 *
 * PARAMETERS:  list - DBListType struct
 *				numItems - number of items
 *				creator - id of creator to be found
 *
 * RETURNED:    item number of matching
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		01/25/04	Initial Revision
 *
 ***********************************************************************/
static Int16 FindListItemByCreator (DBListType *list, UInt16 numItems, UInt32 creator)
{
	UInt16	select = 0;
	UInt16	numMatches = 0;
	UInt16	i;


	for (i = 0; i < numItems; i ++)
	{
		if (list[i].creator == creator)
		{
			select = i;
			numMatches ++;
			if (numMatches > 1)
			{
				select = 0;
				break;
			}
		}
	}

	return (select);
}


/***********************************************************************
 *
 * FUNCTION:    FindListItemByName
 *
 * DESCRIPTION: This routine finds item number in list by name.
 *
 * PARAMETERS:  list - DBListType struct
 *				numItems - number of items
 *				name - string of name to be found
 *
 * RETURNED:    item number of matching
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/08/04	Initial Revision
 *
 ***********************************************************************/
static Int16 FindListItemByName (DBListType *list, UInt16 numItems, Char *name)
{
	UInt16	select = 0;
	UInt16	i;


	for (i = 0; i < numItems; i ++)
	{
		if (StrCompare (list[i].name, name) == 0)
		{
			select = i;
			break;
		}
	}

	return (select);
}


#pragma mark ----
/***********************************************************************
 *
 * FUNCTION:    PluginsFindPlugin
 *
 * DESCRIPTION: This routine find the specified plugin.
 *
 * PARAMETERS:  select
 *
 * RETURNED:    true if found
 *				cardNo
 *				dbID
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		11/25/07	Initial Revision
 *
 ***********************************************************************/
static Boolean PluginsFindPlugin (Int16 select, UInt16 *cardNo, LocalID *dbID)
{
	DmSearchStateType	searchState;
	Boolean				newSearch = true;
	Char				name[dmDBNameLength];
	Boolean				found = false;


	while (true)
	{
		if (DmGetNextDatabaseByTypeCreator (newSearch, &searchState, PsLinkTPlugins, NULL, false, cardNo, dbID) == errNone)
		{
			GetApplName (*cardNo, *dbID, name);
			if (StrCompare (name, g_Pin.names[select]) == 0)
			{
				found = true;
				break;
			}
		}
		else
			break;
		newSearch = false;
	}
	return found;
}


/***********************************************************************
 *
 * FUNCTION:    PluginsDeletePlugin
 *
 * DESCRIPTION: This routine deletes a plugin.
 *
 * PARAMETERS:  select
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		11/25/07	Initial Revision
 *
 ***********************************************************************/
static void PluginsDeletePlugin (Int16 select)
{
	UInt16				cardNo;
	LocalID				dbID;


	if (PluginsFindPlugin (select, &cardNo, &dbID))
	{
		DmDeleteDatabase (cardNo, dbID);
	}
}


/***********************************************************************
 *
 * FUNCTION:    PluginsHelp
 *
 * DESCRIPTION: This routine displays a plugin help.
 *				Resource type : 'tSTR' (strRsc)
 *				Resource ID : 9000 (plinHelpID)
 *
 * PARAMETERS:  select
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		11/28/07	Initial Revision
 *
 ***********************************************************************/
static void PluginsHelp (Int16 select)
{
	DmOpenRef			dbP;
	UInt16				cardNo;
	LocalID				dbID;


	if (PluginsFindPlugin (select, &cardNo, &dbID)
	 && (dbP = DmOpenDatabase (cardNo, dbID, dmModeReadOnly)))
	{
		FrmHelp (plinHelpID);
		DmCloseDatabase (dbP);
	}
}


/***********************************************************************
 *
 * FUNCTION:    PluginsFormInit
 *
 * DESCRIPTION: This routine initializes the Plugins form.
 *
 * PARAMETERS:  frmP - a pointer to the form type structure
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		11/25/07	Initial Revision
 *
 ***********************************************************************/
static void PluginsFormInit (FormPtr frmP)
{
	UInt16	index;


	GetDBList (&g_Pin, TYPE_PLIN, "");
	SetListItems (PluginsPluginsList, g_Pin.names, g_Pin.num);

	index = FrmGetObjectIndex (frmP, PluginsInformationLabel);
	if (g_Pin.num == 0)
		FrmShowObject (frmP, index);
	else
		FrmHideObject (frmP, index);
		
}


/***********************************************************************
 *
 * FUNCTION:    PluginsFormHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the 
 *              "PluginsForm" of this application.
 *
 * PARAMETERS:  eventP  - a pointer to an EventType structure
 *
 * RETURNED:    true if the event has handle and should not be passed
 *              to a higher level handler.
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		11/25/07	Initial Revision
 *
 ***********************************************************************/
static Boolean PluginsFormHandleEvent (EventPtr eventP)
{
	FormPtr		frmP;
	Boolean		handled = false;


	switch (eventP->eType)
	{
		case ctlSelectEvent:
			switch (eventP->data.ctlSelect.controlID)
			{
				case PluginsDoneButton:
					FrmReturnToForm (0);
					handled = true;
					break;

				case PluginsHelpButton:
					PluginsHelp (LstGetSelection ((ListPtr)GetObjectPtr (PluginsPluginsList)));
					handled = true;
					break;

				case PluginsDeleteButton:
					PluginsDeletePlugin (LstGetSelection ((ListPtr)GetObjectPtr (PluginsPluginsList)));
					frmP = FrmGetActiveForm ();
					PluginsFormInit (frmP);
					FrmDrawForm (frmP);
					handled = true;
					break;

				default:
					break;
			}
			break;					

		case frmOpenEvent:
			frmP = FrmGetActiveForm ();
			PluginsFormInit (frmP);
			FrmDrawForm (frmP);
			handled = true;
			break;

		default:
			break;
	}
	
	return handled;
}


#pragma mark ----
/***********************************************************************
 *
 * FUNCTION:    ButtonsFormInit
 *
 * DESCRIPTION: This routine initializes the ButtonsForm form.
 *
 * PARAMETERS:  frmP - a pointer to the form type structure
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		06/11/04	Initial Revision
 *			smb		02/13/05	Modify for a definition of a back-button
 *
 ***********************************************************************/
static void ButtonsFormInit (FormPtr frmP)
{
	if (g_Prefs.backKey.chr == 0)
		FrmShowObject (frmP, FrmGetObjectIndex (frmP, ButtonsBackbuttonNotdefinedLabel));
	else
		FrmShowObject (frmP, FrmGetObjectIndex (frmP, ButtonsBackbuttonDefinedLabel));

	if (g_Prefs.forwardKey.chr == 0)
		FrmShowObject (frmP, FrmGetObjectIndex (frmP, ButtonsForwardbuttonNotdefinedLabel));
	else
		FrmShowObject (frmP, FrmGetObjectIndex (frmP, ButtonsForwardbuttonDefinedLabel));

	FrmSetControlValue (frmP, FrmGetObjectIndex (frmP, ButtonsIgnoreCheckbox), 
						(g_Prefs.ignoreBtn == IsON)? 1 : 0);
}


/***********************************************************************
 *
 * FUNCTION:	ButtonsFormApply
 *
 * DESCRIPTION: This routine applies the changes made in the Buttons form
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		06/11/04	Initial Revision
 *			smb		03/26/05	Support "do nothing" option
 *
 ***********************************************************************/
static void ButtonsFormApply (FormPtr frmP)
{
	if (FrmGetControlValue (frmP, FrmGetObjectIndex (frmP, ButtonsIgnoreCheckbox)))
		g_Prefs.ignoreBtn = IsON;
	else
		g_Prefs.ignoreBtn = IsOFF;
}


/***********************************************************************
 *
 * FUNCTION:	ButtonsFormClearBackkey
 *
 * DESCRIPTION: This routine clear the back-button
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/13/05	Initial Revision
 *
 ***********************************************************************/
static void ButtonsFormClearBackkey (void)
{
	FormPtr		frmP;

	frmP = FrmGetActiveForm ();
	FrmHideObject (frmP, FrmGetObjectIndex (frmP, ButtonsBackbuttonDefinedLabel));
	FrmShowObject (frmP, FrmGetObjectIndex (frmP, ButtonsBackbuttonNotdefinedLabel));
	g_Prefs.backKey.chr = 0;
	g_Prefs.backKey.modifiers = 0;
}


/***********************************************************************
 *
 * FUNCTION:	ButtonsFormDefineBackkey
 *
 * DESCRIPTION: This routine defines the back-button
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/13/05	Initial Revision
 *
 ***********************************************************************/
static void ButtonsFormDefineBackkey (void)
{
	EventType	event;
	FormPtr		frmP;


	// Clear event queues
	do {
		EvtGetEvent (&event, evtNoWait);
	} while (event.eType != nilEvent);

	frmP = FrmGetActiveForm ();
	FrmHideObject (frmP, FrmGetObjectIndex (frmP, ButtonsBackbuttonDefinedLabel));
	FrmHideObject (frmP, FrmGetObjectIndex (frmP, ButtonsBackbuttonNotdefinedLabel));
	FrmShowObject (frmP, FrmGetObjectIndex (frmP, ButtonsBackbuttonHintLabel));

	// Wait a next event
	do {
		EvtGetEvent (&event, evtWaitForever);
	} while (event.eType == nilEvent);

	FrmHideObject (frmP, FrmGetObjectIndex (frmP, ButtonsBackbuttonHintLabel));

	switch (event.eType)
	{
		case keyDownEvent:
			// The pushed key is defined as a back-button
			FrmShowObject (frmP, FrmGetObjectIndex (frmP, ButtonsBackbuttonDefinedLabel));
			g_Prefs.backKey.chr = event.data.keyDown.chr;
			g_Prefs.backKey.modifiers = event.data.keyDown.modifiers;
			break;

		default:
			// No key is defined as a back-button
			ButtonsFormClearBackkey ();
			EvtAddEventToQueue (&event);
			break;
	}
}


/***********************************************************************
 *
 * FUNCTION:	ButtonsFormClearForwardkey
 *
 * DESCRIPTION: This routine clear the forward-button
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/13/05	Initial Revision
 *
 ***********************************************************************/
static void ButtonsFormClearForwardkey (void)
{
	FormPtr		frmP;

	frmP = FrmGetActiveForm ();
	FrmHideObject (frmP, FrmGetObjectIndex (frmP, ButtonsForwardbuttonDefinedLabel));
	FrmShowObject (frmP, FrmGetObjectIndex (frmP, ButtonsForwardbuttonNotdefinedLabel));
	g_Prefs.forwardKey.chr = 0;
	g_Prefs.forwardKey.modifiers = 0;
}


/***********************************************************************
 *
 * FUNCTION:	ButtonsFormDefineForwardkey
 *
 * DESCRIPTION: This routine defines the forward-button
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/13/05	Initial Revision
 *
 ***********************************************************************/
static void ButtonsFormDefineForwardkey (void)
{
	EventType	event;
	FormPtr		frmP;


	// Clear event queues
	do {
		EvtGetEvent (&event, evtNoWait);
	} while (event.eType != nilEvent);

	frmP = FrmGetActiveForm ();
	FrmHideObject (frmP, FrmGetObjectIndex (frmP, ButtonsForwardbuttonDefinedLabel));
	FrmHideObject (frmP, FrmGetObjectIndex (frmP, ButtonsForwardbuttonNotdefinedLabel));
	FrmShowObject (frmP, FrmGetObjectIndex (frmP, ButtonsForwardbuttonHintLabel));

	// Wait a next event
	do {
		EvtGetEvent (&event, evtWaitForever);
	} while (event.eType == nilEvent);

	FrmHideObject (frmP, FrmGetObjectIndex (frmP, ButtonsForwardbuttonHintLabel));

	switch (event.eType)
	{
		case keyDownEvent:
			// The pushed key is defined as a back-button
			FrmShowObject (frmP, FrmGetObjectIndex (frmP, ButtonsForwardbuttonDefinedLabel));
			g_Prefs.forwardKey.chr = event.data.keyDown.chr;
			g_Prefs.forwardKey.modifiers = event.data.keyDown.modifiers;
			break;

		default:
			// No key is defined as a back-button
			ButtonsFormClearForwardkey ();
			EvtAddEventToQueue (&event);
			break;
	}
}


/***********************************************************************
 *
 * FUNCTION:    ButtonsFormHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the 
 *              "ButtonsForm" of this application.
 *
 * PARAMETERS:  eventP  - a pointer to an EventType structure
 *
 * RETURNED:    true if the event has handle and should not be passed
 *              to a higher level handler.
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		06/11/04	Initial Revision
 *			smb		02/13/05	Modify for a definition of a back-button
 *			smb		03/26/05	Modify for a definition of a forward-button
 *
 ***********************************************************************/
static Boolean ButtonsFormHandleEvent (EventPtr eventP)
{
	FormPtr		frmP;
	Boolean		handled = false;


	switch (eventP->eType) 
	{
		case ctlSelectEvent:
			switch (eventP->data.ctlSelect.controlID)
			{
				case ButtonsDoneButton:
					frmP = FrmGetActiveForm ();
					ButtonsFormApply (frmP);
					FrmReturnToForm (0);
					handled = true;
					break;

				case ButtonsBackbuttonDefineButton:
					ButtonsFormDefineBackkey ();
					handled = true;
					break;

				case ButtonsBackbuttonClearButton:
					ButtonsFormClearBackkey ();
					handled = true;
					break;

				case ButtonsForwardbuttonDefineButton:
					ButtonsFormDefineForwardkey ();
					handled = true;
					break;

				case ButtonsForwardbuttonClearButton:
					ButtonsFormClearForwardkey ();
					handled = true;
					break;
			}
			break;					

		case frmOpenEvent:
			frmP = FrmGetActiveForm ();
			ButtonsFormInit (frmP);
			FrmDrawForm (frmP);
			handled = true;
			break;

		default:
			break;
	}
	
	return handled;
}


#pragma mark ----
/***********************************************************************
 *
 * FUNCTION:    SelectMoreAppList
 *
 * DESCRIPTION:	This routine handles to popup and select 
 *				an application list in MoreForm.
 *
 * PARAMETERS:	nothing
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/10/04	Initial Revision
 *			smb		12/18/04	Support database link and database prefix
 *
 ***********************************************************************/
static void SelectMoreAppList (FormPtr frmP)
{
	UInt16		selApp, selDb;
	Char		str[MAXLEN_PREFIX + 1];


	// Select application
	SelectList (MoreAppPopTrigger, MoreAppList, NULL);
	selApp = LstGetSelection ((ListPtr)GetObjectPtr (MoreAppList));

	// Choose database automatically
	selDb = FindListItemByCreator (g_DB.list, g_DB.num, DbCreator(g_App.list[selApp].creator));
	SetTriggerLabel (MoreDBPopTrigger, MoreDBList, selDb, NULL);

	// Database link?
	FrmSetControlValue (frmP, FrmGetObjectIndex (frmP, MoreDBlinkCheckbox), 0);
	if (selDb == 0)
	{
		StrCopy (str, DbPrefix(g_App.list[selApp].creator));
		SetFieldString (MoreDBPrefixField, str);
	}
	else
	{
		SetFieldString (MoreDBPrefixField, "");
	}
}


/***********************************************************************
 *
 * FUNCTION:    MoreFormInit
 *
 * DESCRIPTION: This routine initializes the MoreForm form.
 *
 * PARAMETERS:  frmP - pointer to the MoreForm form.
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/08/04	Initial Revision
 *			smb		12/18/04	Support database link and database prefix
 *
 ***********************************************************************/
static void MoreFormInit (FormPtr frmP)
{
	UInt16		select;


	g_AppNum = FrmGetObjectId (frmP, FrmGetControlGroupSelection (frmP, MoreGroupID))
	         - More1PushButton + NUM_DEF_LINK_APPS;

	SetListItems (MoreAppList, g_App.names, g_App.num);
	select = FindListItemByCreator (g_App.list, g_App.num, g_Prefs.creator[g_AppNum]);
	SetTriggerLabel (MoreAppPopTrigger, MoreAppList, select, NULL);

	SetListItems (MoreDBList, g_DB.names, g_DB.num);
	select = FindListItemByName (g_DB.list, g_DB.num, g_Prefs.dbName[g_AppNum]);
	SetTriggerLabel (MoreDBPopTrigger, MoreDBList, select, NULL);

	SetFieldString (MoreStartField, g_Prefs.tag[g_AppNum][START]);
	SetFieldString (MoreEndField, g_Prefs.tag[g_AppNum][END]);

	FrmSetControlValue (frmP, FrmGetObjectIndex (frmP, MoreDBlinkCheckbox), (g_Prefs.dbLink[g_AppNum])? 1 : 0);
	SetFieldString (MoreDBPrefixField, g_Prefs.dbPrefix[g_AppNum]);
}


/***********************************************************************
 *
 * FUNCTION:	MoreFormApply
 *
 * DESCRIPTION: This routine applies the changes made in the More form
 *
 * PARAMETERS:  frmP - pointer to the MoreForm form.
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/08/04	Initial Revision
 *			smb		12/18/04	Support database link and database prefix
 *
 ***********************************************************************/
static void MoreFormApply (FormPtr frmP)
{
	ListPtr		listP;
	UInt16		select;


	// Application creator IDs
	listP = (ListPtr)GetObjectPtr (MoreAppList);
	select = LstGetSelection (listP);
	g_Prefs.creator[g_AppNum] = g_App.list[select].creator;

	// Database name
	listP = (ListPtr)GetObjectPtr (MoreDBList);
	select = LstGetSelection (listP);
	if (select == 0)
		StrCopy (g_Prefs.dbName[g_AppNum], "");
	else
		StrCopy (g_Prefs.dbName[g_AppNum], g_DB.list[select].name);

	// Tags
	GetFieldString (MoreStartField, MAXLEN_TAG, g_Prefs.tag[g_AppNum][START]);
	GetFieldString (MoreEndField, MAXLEN_TAG, g_Prefs.tag[g_AppNum][END]);

	// Set a application to "none" if no tag defined
	if (*g_Prefs.tag[g_AppNum][START] == NULL
	 || *g_Prefs.tag[g_AppNum][END] == NULL)
		g_Prefs.creator[g_AppNum] = LINKAPP_NONE_ID;

	// DB link and DB name prefix
	if (FrmGetControlValue (frmP, FrmGetObjectIndex (frmP, MoreDBlinkCheckbox)))
		g_Prefs.dbLink[g_AppNum] = true;
	else
		g_Prefs.dbLink[g_AppNum] = false;
	GetFieldString (MoreDBPrefixField, MAXLEN_PREFIX, g_Prefs.dbPrefix[g_AppNum]);
}


/***********************************************************************
 *
 * FUNCTION:    MoreFormHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the 
 *              "MoreForm" of this application.
 *
 * PARAMETERS:  eventP  - a pointer to an EventType structure
 *
 * RETURNED:    true if the event has handle and should not be passed
 *              to a higher level handler.
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/08/04	Initial Revision
 *
 ***********************************************************************/
static Boolean MoreFormHandleEvent (EventPtr eventP)
{
	FormPtr		frmP;
	Boolean		handled = false;


	switch (eventP->eType) 
	{
		case ctlSelectEvent:
			switch (eventP->data.ctlSelect.controlID)
			{
				case MoreDoneButton:
					frmP = FrmGetActiveForm ();
					MoreFormApply (frmP);
					FrmReturnToForm (0);
					handled = true;
					break;

				case More1PushButton:
				case More2PushButton:
				case More3PushButton:
				case More4PushButton:
				case More5PushButton:
					frmP = FrmGetActiveForm ();
					MoreFormApply (frmP);
					MoreFormInit (frmP);
					handled = true;
					break;

				case MoreAppPopTrigger:
					frmP = FrmGetActiveForm ();
					SelectMoreAppList (frmP);
					handled = true;
					break;					

				/*
				case MoreDBPopTrigger:
					SelectList (MoreDBPopTrigger, MoreDBList, NULL);
					handled	= true;
					break;					
				*/
			}
			break;					

		case frmOpenEvent:
			frmP = FrmGetActiveForm ();
			FrmSetControlGroupSelection (frmP, MoreGroupID, More1PushButton);
			MoreFormInit (frmP);
			FrmDrawForm (frmP);
			handled = true;
			break;

		default:
			break;
	}
	
	return handled;
}


#pragma mark ----
/***********************************************************************
 *
 * FUNCTION:    MainFormSetOptions
 *
 * DESCRIPTION: This routine sets options in OptionsForm.
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		03/26/05	Initial Revision
 *
 ***********************************************************************/
static void MainFormSetOptions (void)
{
	FormPtr		frmP, orgFrmP;
	FieldPtr	fldP;


	orgFrmP = FrmGetActiveForm ();

	frmP = FrmInitForm (OptionsForm);
	FrmSetActiveForm (frmP);
	FrmSetControlValue (frmP, FrmGetObjectIndex (frmP, OptionsMemoWikiNameCheckbox),
						 (g_Prefs.memoWikiName == IsON)? 1 : 0);
	FrmSetControlValue (frmP, FrmGetObjectIndex (frmP, OptionsToDoCompletedCheckbox),
						 (g_Prefs.todoCompleted == IsON)? 1 : 0);
	FrmSetControlValue (frmP, FrmGetObjectIndex (frmP, OptionsCasesensitiveCheckbox),
						 (g_Prefs.caseSensitive == IsON)? 1 : 0);
	FrmSetControlValue (frmP, FrmGetObjectIndex (frmP, OptionsPartoftitleCheckbox),
						 (g_Prefs.partOfTitle == IsON)? 1 : 0);
	if (g_Prefs.linkChrBookmark)
	{
		fldP = (FieldPtr)GetObjectPtr (OptionsBookmarkField);
		FldInsert (fldP, &g_Prefs.linkChrBookmark, 1);
	}
	if (g_Prefs.linkChrDBname)
	{
		fldP = (FieldPtr)GetObjectPtr (OptionsDBlinkField);
		FldInsert (fldP, &g_Prefs.linkChrDBname, 1);
	}

	FrmDoDialog (frmP);

	g_Prefs.memoWikiName = FrmGetControlValue (frmP, 
						FrmGetObjectIndex (frmP, OptionsMemoWikiNameCheckbox));
	g_Prefs.todoCompleted = FrmGetControlValue (frmP, 
						FrmGetObjectIndex (frmP, OptionsToDoCompletedCheckbox));
	g_Prefs.caseSensitive = FrmGetControlValue (frmP, 
						FrmGetObjectIndex (frmP, OptionsCasesensitiveCheckbox));
	g_Prefs.partOfTitle = FrmGetControlValue (frmP, 
						FrmGetObjectIndex (frmP, OptionsPartoftitleCheckbox));
	GetFieldString (OptionsBookmarkField, 1, &g_Prefs.linkChrBookmark);
	GetFieldString (OptionsDBlinkField, 1, &g_Prefs.linkChrDBname);

	FrmDeleteForm (frmP);
	FrmSetActiveForm (orgFrmP);
}


/***********************************************************************
 *
 * FUNCTION:    MainFormDoCommand
 *
 * DESCRIPTION: This routine performs the menu command specified.
 *
 * PARAMETERS:  command  - menu item id
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/29/03	Initial Revision
 *
 ***********************************************************************/
static Boolean MainFormDoCommand (UInt16 command)
{
	Boolean handled = false;
	FormPtr frmP;

	switch (command)
	{
		case MainOptionsOptions:
			MainFormSetOptions ();
			handled = true;
			break;

		case MainOptionsButtons:
			FrmPopupForm (ButtonsForm);
			handled = true;
			break;

		case MainOptionsPlugins:
			FrmPopupForm (PluginsForm);
			handled = true;
			break;

		case MainOptionsAbout:
			MenuEraseStatus (0);			// Clear the menu status from the display.
			frmP = FrmInitForm (AboutForm);
			FrmDoDialog (frmP);				// Display the About Box.
			FrmDeleteForm (frmP);
			handled = true;
			break;
	}
	
	return handled;
}


/***********************************************************************
 *
 * FUNCTION:    MainFormInit
 *
 * DESCRIPTION: This routine initializes the MainForm form.
 *
 * PARAMETERS:  frmP - pointer to the MainForm form.
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/03	Initial Revision
 *			smb		02/07/04	Modify to support ToDo application
 *
 ***********************************************************************/
static void MainFormInit (FormPtr frmP)
{
	UInt16			select;
	Int16			i;


	// Set the control value that indicates option value.
	FrmSetControlGroupSelection (frmP, MainGroupID,
			 (g_Prefs.state == IsON)? MainONPushButton : MainOFFPushButton);

	FrmSetControlValue (frmP, FrmGetObjectIndex (frmP, MainSingleTapCheckbox),
			 (g_Prefs.singleTap == IsON)? 1 : 0);

	FrmSetControlValue (frmP, FrmGetObjectIndex (frmP, MainAutoEnableCheckbox),
			 (g_Prefs.autoEnble == IsON)? 1 : 0);

	// Set the field string for link tag.
	SetFieldString (MainMemoStartField, g_Prefs.tag[MEMO][START]);
	SetFieldString (MainMemoEndField, g_Prefs.tag[MEMO][END]);
	SetFieldString (MainToDoStartField, g_Prefs.tag[TODO][START]);
	SetFieldString (MainToDoEndField, g_Prefs.tag[TODO][END]);
	SetFieldString (MainAddressStartField, g_Prefs.tag[ADDRESS][START]);
	SetFieldString (MainAddressEndField, g_Prefs.tag[ADDRESS][END]);

	// Set the popup list for link application.
	for (i = 0; i < NUM_DEF_LINK_APPS; i ++)
	{
		SetListItems (ListID(i), g_App.names, g_App.num);
		select = FindListItemByCreator (g_App.list, g_App.num, g_Prefs.creator[i]);
		SetTriggerLabel (CtrlID(i), ListID(i), select, WIDTH_POPTRIGGER);
	}

	// Set the popup and the list for a preferences panel
	if (g_PnlListAvailable)
	{
		FrmHideObject (frmP, FrmGetObjectIndex (frmP, MainPanelNameLabel));

		SetListItems (MainPanelList, g_Pnl.names, g_Pnl.num);
		select = FindListItemByCreator (g_Pnl.list, g_Pnl.num, appFileCreator);
		SetTriggerLabel (MainPanelPopTrigger, MainPanelList, select, NULL);
	}
	else
	{
		FrmHideObject (frmP, FrmGetObjectIndex (frmP, MainPanelPopTrigger));
	}

	EnableNavigation (frmP);
}


/***********************************************************************
 *
 * FUNCTION:	MainFormApply
 *
 * DESCRIPTION: This routine applies the changes made in the Main form
 *
 * PARAMETERS:  frmP - pointer to the MainForm form.
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/03	Initial Revision
 *			smb		02/07/04	Modify to support ToDo application
 *
 ***********************************************************************/
static void MainFormApply (FormPtr frmP)
{
	ListPtr				listP;
	UInt16				select;
	Int16				i;


	// Get some option value.
	if (FrmGetControlValue (frmP, FrmGetObjectIndex (frmP, MainONPushButton)))
		g_Prefs.state = IsON;
	else
		g_Prefs.state = IsOFF;

	if (FrmGetControlValue (frmP, FrmGetObjectIndex (frmP, MainSingleTapCheckbox)))
		g_Prefs.singleTap = IsON;
	else
		g_Prefs.singleTap = IsOFF;

	if (FrmGetControlValue (frmP, FrmGetObjectIndex (frmP, MainAutoEnableCheckbox)))
		g_Prefs.autoEnble = IsON;
	else
		g_Prefs.autoEnble = IsOFF;

	// Get creators of selected application
	for (i = 0; i < NUM_DEF_LINK_APPS; i ++)
	{
		listP = (ListPtr)GetObjectPtr (ListID(i));
		select = LstGetSelection (listP);
		g_Prefs.creator[i] = g_App.list[select].creator;
	}

	// Get tags
	GetFieldString (MainMemoStartField, MAXLEN_TAG, g_Prefs.tag[MEMO][START]);
	GetFieldString (MainMemoEndField, MAXLEN_TAG, g_Prefs.tag[MEMO][END]);
	GetFieldString (MainToDoStartField, MAXLEN_TAG, g_Prefs.tag[TODO][START]);
	GetFieldString (MainToDoEndField, MAXLEN_TAG, g_Prefs.tag[TODO][END]);
	GetFieldString (MainAddressStartField, MAXLEN_TAG, g_Prefs.tag[ADDRESS][START]);
	GetFieldString (MainAddressEndField, MAXLEN_TAG, g_Prefs.tag[ADDRESS][END]);

	// Set the application "none" if no tag is written in the field
	if (*g_Prefs.tag[MEMO][START] == NULL || *g_Prefs.tag[MEMO][END] == NULL)
		g_Prefs.creator[MEMO] = LINKAPP_NONE_ID;
	if (*g_Prefs.tag[TODO][START] == NULL || *g_Prefs.tag[TODO][END] == NULL)
		g_Prefs.creator[TODO] = LINKAPP_NONE_ID;
	if (*g_Prefs.tag[ADDRESS][START] == NULL || *g_Prefs.tag[ADDRESS][END] == NULL)
		g_Prefs.creator[ADDRESS] = LINKAPP_NONE_ID;
}


/***********************************************************************
 *
 * FUNCTION:    MainFormHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the 
 *              "MainForm" of this application.
 *
 * PARAMETERS:  eventP  - a pointer to an EventType structure
 *
 * RETURNED:    true if the event has handle and should not be passed
 *              to a higher level handler.
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/03	Initial Revision
 *
 ***********************************************************************/
static Boolean MainFormHandleEvent (EventPtr eventP)
{
	FormPtr		frmP;
	EventType	event;
	Boolean		handled = false;


	switch (eventP->eType) 
	{
		case ctlSelectEvent:
			switch (eventP->data.ctlSelect.controlID)
			{
				case MainDoneButton:
					event.eType = appStopEvent;
					EvtAddEventToQueue (&event);
					handled = true;
					break;					

				case MainMoreButton:
					FrmPopupForm (MoreForm);
					handled = true;
					break;

				case MainPanelPopTrigger:
					handled = SelectPanel ();
					break;

				case MainMemoPopTrigger:
				case MainAddressPopTrigger:
				case MainDatePopTrigger:
				case MainWebPopTrigger:
				case MainMailPopTrigger:
					SelectList (eventP->data.ctlSelect.controlID,
					 Ctl2List(eventP->data.ctlSelect.controlID), WIDTH_POPTRIGGER);
					handled = true;
					break;					
			}
			break;					

		case menuEvent:
			return MainFormDoCommand (eventP->data.menu.itemID);

		case frmOpenEvent:
			frmP = FrmGetActiveForm ();
			MainFormInit (frmP);
			FrmDrawForm (frmP);
			handled = true;
			break;

		case frmCloseEvent:
			frmP = FrmGetActiveForm ();
			MainFormApply (frmP);
			//handled = true;
			break;

		/*
		case frmUpdateEvent:
			// To do any custom drawing here, first call FrmDrawForm(), then do your
			// drawing, and then set handled to true.
			break;
		*/

		default:
			break;
	}
	
	return handled;
}


#pragma mark ----
/***********************************************************************
 *
 * FUNCTION:    AppHandleEvent
 *
 * DESCRIPTION: This routine loads form resources and set the event
 *              handler for the form loaded.
 *
 * PARAMETERS:  event  - a pointer to an EventType structure
 *
 * RETURNED:    true if the event has handle and should not be passed
 *              to a higher level handler.
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/29/03	Initial Revision
 *
 ***********************************************************************/
static Boolean AppHandleEvent (EventPtr eventP)
{
	UInt16 formId;
	FormPtr frmP;

	if (eventP->eType == frmLoadEvent)
	{
		// Load the form resource.
		formId = eventP->data.frmLoad.formID;
		frmP = FrmInitForm (formId);
		FrmSetActiveForm (frmP);

		// Set the event handler for the form.  The handler of the currently
		// active form is called by FrmHandleEvent each time is receives an
		// event.
		switch (formId)
		{
			case MainForm:
				FrmSetEventHandler(frmP, MainFormHandleEvent);
				break;

			case MoreForm:
				FrmSetEventHandler(frmP, MoreFormHandleEvent);
				break;

			case ButtonsForm:
				FrmSetEventHandler(frmP, ButtonsFormHandleEvent);
				break;

			case PluginsForm:
				FrmSetEventHandler(frmP, PluginsFormHandleEvent);
				break;

			default:
//				ErrFatalDisplay("Invalid Form Load Event");
				break;

		}
		return true;
	}
	
	return false;
}


/***********************************************************************
 *
 * FUNCTION:    AppEventLoop
 *
 * DESCRIPTION: This routine is the event loop for the application.  
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/29/03	Initial Revision
 *
 ***********************************************************************/
static void AppEventLoop (void)
{
	UInt16		error;
	EventType	event;

	do {
		EvtGetEvent(&event, evtWaitForever);

		if (! SysHandleEvent(&event))
			if (! MenuHandleEvent(0, &event, &error))
				if (! AppHandleEvent(&event))
					FrmDispatchEvent(&event);

	} while (event.eType != appStopEvent);
}


/***********************************************************************
 *
 * FUNCTION:     AppStart
 *
 * DESCRIPTION:  Get the current application's preferences.
 *
 * PARAMETERS:   nothing
 *
 * RETURNED:     Err value 0 if nothing went wrong
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/03	Initial Revision
 *			smb		02/08/04	Modify to support more applications.
 *
 ***********************************************************************/
static Err AppStart (void)
{
	Err		err;
	Char	str[dmDBNameLength];	


	if ((err = RomVersionCompatible (kOurMinVersion)) != errNone)
		return err;
	if ((err = NotificationAvailable ()) != errNone)
		return err;

	Disable ();

	GetPreferences (&g_Prefs);

	// Initialize the application list.
	SysStringByIndex (dbNameStringList, NONE_NAME, str, dmDBNameLength); 
	GetDBList (&g_App, TYPE_APPL, str);

	// Initialize the database list.
	SysStringByIndex (dbNameStringList, DEF_NAME, str, dmDBNameLength); 
	GetDBList (&g_DB, TYPE_DATA, str);

	// Initialize the list for a preferences panel
	GetDBList (&g_Pnl, TYPE_PANL, "");

	g_FepMode = TsmSetFepMode (NULL, tsmFepModeOff);

	return errNone;
}


/***********************************************************************
 *
 * FUNCTION:    AppStop
 *
 * DESCRIPTION: Save the current state of the application.
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/03	Initial Revision
 *			smb		02/08/04	Modify to support more applications.
 *
 ***********************************************************************/
static void AppStop (void)
{
	// Close all the open forms.
	FrmCloseAllForms ();

	SetPreferences (&g_Prefs);

	TsmSetFepMode (NULL, g_FepMode);

	// Free chunks for lists
	FreeDBList (g_App);
	FreeDBList (g_DB);
	FreeDBList (g_Pnl);

	if (g_Prefs.state == IsON)
		Enable (&g_Prefs);
}


/***********************************************************************
 *
 * FUNCTION:    PalmMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 *
 * PARAMETERS:  cmd - word value specifying the launch code. 
 *              cmdPB - pointer to a structure that is associated with the launch code. 
 *              launchFlags -  word value providing extra information about the launch.
 *
 * RETURNED:    Result of launch
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/29/03	Initial Revision
 *
 ***********************************************************************/
static UInt32 PalmMain (UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	Err						err;
	PsLinkPreferenceType	prefs;


	switch (cmd)
	{
		case sysAppLaunchCmdReturnFromPanel:
		case sysAppLaunchCmdNormalLaunch:
			g_PnlListAvailable = true;
		case sysAppLaunchCmdPanelCalledFromApp:
			err = AppStart();
			if (err)
				return (err);
				
			FrmGotoForm (MainForm);
			AppEventLoop();
			AppStop();
			break;

		case sysAppLaunchCmdSystemReset:
			err = NotificationAvailable ();
			if (err)
				return (err);
			GetPreferences (&prefs);
			if (prefs.state == IsON)
			{
				if (prefs.autoEnble == IsON)
					Enable (&prefs);
				else
				{
					Disable ();
					prefs.state  = IsOFF;
				}
				SetPreferences (&prefs);
			}
			((SysNotifyParamType *)cmdPBP)->handled = true;
			break;

		case sysAppLaunchCmdNotify:
			SysNotifyParamType	*notifyParam = (SysNotifyParamType *)cmdPBP;
			switch (notifyParam->notifyType)
			{
				case sysNotifyEventDequeuedEvent:
					DequeuedEventHandle (notifyParam);
					break;

				case appFileCreator:
					CreatorEventHandle (notifyParam);
					break;
			}
			break;

		default:
			break;

	}
	
	return errNone;
}


/***********************************************************************
 *
 * FUNCTION:    PilotMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 *
 * PARAMETERS:  cmd - word value specifying the launch code. 
 *              cmdPB - pointer to a structure that is associated with the launch code. 
 *              launchFlags -  word value providing extra information about the launch.
 * RETURNED:    Result of launch
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/29/03	Initial Revision
 *
 ***********************************************************************/
UInt32 PilotMain (UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	return PalmMain (cmd, cmdPBP, launchFlags);
}
