/******************************************************************************
 *
 * Copyright (c) 2004 PalmSpace/Sembe.
 *				All rights reserved.
 *
 * File: PsLink.h
 *
 * History:
 *		Dec 29, 2003	Created by sembe
 *
 *****************************************************************************/

#ifndef _PSLINK_H_
#define _PSLINK_H_


#include <PalmOS.h>
#include <PalmUtils.h>
#include <SonyCLIE.h>
#include <CharShiftJIS.h>
#include <palmOne_68K.h>
#include <Hs.h>
#include <PalmOSGlue.h>


/***********************************************************************
 *
 *	Internal Constants
 *
 ***********************************************************************/
#define appFileCreator			'PsLi'
#define appVersionNum			0x01
#define appPrefID				0x00
#define appPrefVersionNum		0x05

// Define feature number
#define FeatureState			0x00
#define FeatureSelect			0x01
#define FeatureSelStart			0x02
#define FeatureSelEnd			0x03
#define FeatureNavPressed		0x04
#define FeatureBackKey			0x05
#define FeatureForwardKey		0x06

// Plugin help ID
#define plinHelpID				9000

// Define the minimum OS version we support (5.0 for now).
#define kOurMinVersion			sysMakeROMVersion (5,0,0,0,0)

#define TAG_MEMO_START			"["
#define TAG_MEMO_END			"]"
#define TAG_TODO_START			"["
#define TAG_TODO_END			"]t"
#define TAG_ADDRESS_START		"["
#define TAG_ADDRESS_END			"]a"
#define TAG_MORE_START			"["
#define TAG_MORE_END			"]%u"
#define TAG_URLSTART			"http://"

#define LINKCHR_BOOKMARK		':'
#define LINKCHR_DBNAME			'*'

#define MAXLEN_TAG				4
#define MAXLEN_PREFIX			10
#define MAXLEN_LINKSTRING		200
#define WIDTH_POPTRIGGER		50

#define LINKAPP_NONE_ID			0
#define NO_RECORD				0xffff
#define NO_UNIQUEID				0xffffffff

enum DBNAME_STRING_NUM {
	NONE_NAME = 0,
	DEF_NAME
};

enum DBTYPE_NUM {
	TYPE_APPL,
	TYPE_PANL,
	TYPE_DATA,
	TYPE_PLIN
};

enum STATUS_NUM {
	IsOFF = 0,
	IsON = 1
};

enum DEF_APPS_NUM {
	MEMO = 0,
	TODO,
	ADDRESS,
	DATE,
	WEB,
	MAIL,
	NUM_DEF_LINK_APPS
};

enum MORE_APPS_NUM {
	MORE1 = NUM_DEF_LINK_APPS,
	MORE2,
	MORE3,
	MORE4,
	MORE5,
	NUM_LINK_APPS,
	NONE = 0xff
};

#define NUM_MORE_LINK_APPS		NUM_LINK_APPS - NUM_DEF_LINK_APPS

enum TAG_POS_NUM {
	START = 0,
	END
};

#define LENGTH(a) 	(sizeof (a) / sizeof (a)[0])
#define Swap16(n)	(((((UInt16)n) << 8) & 0xFF00) | ((((UInt16)n) >> 8) & 0x00FF))

#define IsData(t)	((t) == 'DATA' || (t) == 'data' || (t) == 'Data' || \
					 (t) == 'TEXt' || 	/* Doc */ \
					 (t) == 'BDOC' ||	/* Wordsmith */ \
					 (t) == 'BMD2' ||	/* BugMe! */ \
					 (t) == 'CMMa' ||	/* CLIEMemo */ \
					 (t) == 'Coog' ||	/* ShadowPlan */ \
					 (t) == 'DATO' ||	/* Bonsai */ \
					 (t) == 'DatV' ||	/* Akiba Map */ \
					 (t) == 'DB00' ||	/* Pilot-DB */ \
					 (t) == 'Foto' ||	/* Palm Photos */ \
					 (t) == 'IMVS' ||	/* PG Pocket */ \
					 (t) == 'JfDb' ||	/* JFile */ \
					 (t) == 'JfD5' ||	/* JFile5 */ \
					 (t) == 'List' ||	/* HandyLists */ \
					 (t) == 'Mdb1' ||	/* MobileDB */ \
					 (t) == 'MSGS' ||	/* BugMe!NotePad */ \
					 (t) == 'Outl' ||	/* ThoughtManager */ \
					 (t) == 'RCH1' || (t) == 'SSPR' || (t) == 'STG1' ||	/* Documents To Go (Word/Slide/Sheet)*/ \
					 (t) == 'ToGo' ||	/* iSilo */ \
					 (t) == 'vIMG')		/* FireViewer */

#define DbCreator(c)	(((c) == 'KStv')? 'View' :	/* TinyViewer */ \
						 ((c) == 'Spec')? 'View' :	/* Spec */ \
						 ((c) == 'PIco')? 'JFil' : 	/* Pico */ \
						 ((c) == 'PoLa')? 'IMVS' : 	/* Polar */ \
						 (c))

#define DbPrefix(c)	(((c) == 'NOtl')? "NO_" :		/* Bonsai */ \
					 ((c) == '2Shp')? "HS2_" :		/* HandyShopper */	\
					 ((c) == 'PMTn')? "PMTxl" :		/* PMT xl */ \
					 ((c) == 'lbPG')? "lbPG-" :		/* Progect */ \
					 ((c) == 'Coog')? "ShadP-" :	/* ShadowPlan */ \
					 ((c) == 'THNK')? "TD2_" :		/* ThinkDB */ \
					 "")


#define IsSJIS(c)	(((UInt8)(c) > 0x80 && (UInt8)(c) < 0xA0) || \
					 ((UInt8)(c) > 0xDF && (UInt8)(c) < 0xFF))
#define IsUpper(c)	((c) >= 'A' && (c) <= 'Z')
#define IsLower(c)	((c) >= 'a' && (c) <= 'z')
#define IsAlpha(c)	((c) >= chrCapital_A && (c) <= chrCapital_Z || (c) >= chrSmall_A && (c) <= chrSmall_Z)
#define IsDigit(c)	((c) >= chrDigitZero && (c) <= chrDigitNine)
#define IsSign(c)	((c) >= chrExclamationMark && (c) <= chrSolidus && (c) != chrQuotationMark || \
	(c) >= chrColon && (c) <= chrCommercialAt && (c) != chrLessThanSign && (c) != chrGreaterThanSign || \
	(c) == chrLowLine || (c) == chrTilde && (c))
#define IsURL(c)	(IsAlpha(c) || IsDigit(c) || IsSign(c))

// 5way-navi
#define NavSelectDirectionPressed(eventP, nav) \
	(((eventP)->data.keyDown.keyCode & (navBitsAll)) == (navBit ## nav | navBitSelect))

#define IsFiveWayNavPalmEvent2(eventP) \
	((((eventP)->data.keyDown.chr == vchrNavChange)) \
	&& (((eventP)->data.keyDown.keyCode & (navBitsAll | navChangeBitsAll)) != 0))

#define IsFiveWayNavPressed2(eventP) \
	(((eventP)->data.keyDown.keyCode == (navBitLeft | navChangeLeft)) || \
	((eventP)->data.keyDown.keyCode == (navBitRight | navChangeRight)) || \
	((eventP)->data.keyDown.keyCode == (navBitSelect | navChangeSelect)))


#endif //_PSLINK_H_
