/******************************************************************************
 *
 * Copyright (c) 2004 Sembe / PalmSpace, All rights reserved.
 *
 * File: PsLinkDA.c
 *
 *****************************************************************************/

#include "PsLink.h"
#include "PsLinkLib.h"
#include "PsLinkDArsc.h"


/***********************************************************************
 *
 *	Entry Points
 *
 ***********************************************************************/
void DAMain (void);


/***********************************************************************
 *
 *	Internal Constants
 *
 ***********************************************************************/
#define FeatureDA		0xff

#define IsSeparator(c)	((c) >= chrComma && (c) <= chrSolidus || \
						 (c) == chrSpace || (c) == chrColon)

#define moveAmt(c)		(((c) == winUp)? -1 : ((c) == winDown)? +1 : 0)


/***********************************************************************
 *
 *	Internal Structures
 *
 ***********************************************************************/
typedef struct {
	FieldPtr	fldP;
	UInt16		start;
	UInt16		end;
	Char		str[MAXLEN_LINKSTRING + MAXLEN_TAG * 2 + 1];
	UInt16		appl;
	Char		appName[NUM_LINK_APPS - 2][dmDBNameLength];
	Char		*array[NUM_LINK_APPS - 2];
	UInt16		appNum[NUM_LINK_APPS - 2];
	Boolean		jogPushed;
} ParamType;


/***********************************************************************
 *
 *	Internal Functions
 *
 ***********************************************************************/


/***********************************************************************
 *
 * FUNCTION:    GetFocusFieldPtr
 *
 * DESCRIPTION: This routine returns a pointer to a form type object in focus.
 *
 * PARAMETERS:	nothing
 *
 * RETURNED:    FieldPtr
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static FieldPtr GetFocusFieldPtr (void)
{
	FormPtr	frmP;
	UInt16	focus;

	if (FrmGetActiveFormID () == 0
	 || (frmP = FrmGetActiveForm ()) == NULL
     || (focus = FrmGetFocus (frmP)) == noFocus)
	{
		return NULL;
    }

	switch (FrmGetObjectType (frmP, focus))
	{
		case frmFieldObj:
			return (FieldPtr)FrmGetObjectPtr (frmP, focus);
		case frmTableObj:
			return TblGetCurrentField ((TablePtr)FrmGetObjectPtr (frmP, focus));
		default:
			return NULL;
	}
}


/***********************************************************************
 *
 * FUNCTION:    SelectDate
 *
 * DESCRIPTION: This routine inserts the week number of the specified date.
 *
 * PARAMETERS:	nothing
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		06/06/04	Initial Revision
 *
 ***********************************************************************/
static Boolean SelectDate (DateType *date)
{
	Char		*title;
	MemHandle	titleH;
	Int16		month, day, year;
	Boolean		select = false;


	day = date->day;
	month = date->month;
	year = date->year + firstYear;

	// Get the title for the date picker dialog box.
	titleH = DmGetResource (strRsc, SelectDateString);
	title = MemHandleLock (titleH);

	if (SelectDay (selectDayByDay, &month, &day, &year, title))
	{
		date->day = day;
		date->month = month;
		date->year = year - firstYear;
		select = true;
	}

	MemHandleUnlock (titleH);

	return select;
}


/***********************************************************************
 *
 * FUNCTION:	DateTimeTo6digit
 *
 * DESCRIPTION:	This routine inserts current date or time.
 *
 * PARAMETERS:	select - date or time
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb	02/16/03	Initial Revision
 *
 ***********************************************************************/
static void DateTimeTo6digit (Char *string)
{
	Char	timString[timeStringLength];
	UInt16	start, pos, len, i, j;


	start = pos = 0;
	len = StrLen (string);
	for (i = 0; i <= len; i ++)
	{
		if (IsSeparator(string[i]) || i == len)
		{
			if (i - start < 2)
			{
				timString[pos] = '0';
				pos ++;
			}
			for (j = start; j <= i; j ++)
			{
				timString[pos] = string[j];
				pos ++;
			}
			start = i + 1;
		}
	}
	StrCopy (string, timString);
}


/***********************************************************************
 *
 * FUNCTION:    GetDstString
 *
 * DESCRIPTION: This routine gets string to be set to the original field.
 *
 * PARAMETERS:	paramP - a pointer to an ParamType structure
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/18/04	Initial Revision
 *
 ***********************************************************************/
static void GetDstString (ParamType *paramP)
{
	PsLinkPreferenceType	prefs;
	SystemPreferencesType	sysPrefs;
	DateType	date;
	Char		str[MAXLEN_LINKSTRING];


	if (paramP->appl == NONE)
		return;
	else if (paramP->appl == DATE)
	{	// Select date
		DateSecondsToDate (TimGetSeconds(), &date);
		if (!SelectDate (&date))
			return;
		
		// Convert the date format using system preference
		PrefGetPreferences (&sysPrefs);
		DateToAscii (date.month, date.day, date.year + firstYear,
		             sysPrefs.dateFormat, paramP->str);
		DateTimeTo6digit (paramP->str);
	}
	else
	{
		// Get preferences of PsLink
		GetPreferences (&prefs);

		// Store tags
		StrCopy (str, paramP->str);
		StrPrintF (paramP->str, "%s%s%s", prefs.tag[paramP->appl][START],
		                             str, prefs.tag[paramP->appl][END]);
	}
}


/***********************************************************************
 *
 * FUNCTION:    DAGetFieldString
 *
 * DESCRIPTION: This routine gets selected string from focused field.
 *
 * PARAMETERS:	paramP
 *
 * RETURNED:	true if the field is usable and editable
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/18/04	Initial Revision
 *
 ***********************************************************************/
static Boolean DAGetFieldString (ParamType *paramP)
{
	FieldAttrType	attr;
	Char			*str;
	UInt16			len;


	paramP->fldP = GetFocusFieldPtr ();
	if (paramP->fldP == NULL)
		return false;
	
	FldGetAttributes (paramP->fldP, &attr);
	if (!attr.editable)
		return false;

	FldGetSelection (paramP->fldP, &paramP->start, &paramP->end);
	if (paramP->end == paramP->start)
	{
		StrCopy (paramP->str, "");
	}
	else
	{
		if (paramP->end - paramP->start > MAXLEN_LINKSTRING)
			paramP->end = paramP->start + MAXLEN_LINKSTRING;
		str = FldGetTextPtr (paramP->fldP);
		len = paramP->end - paramP->start;
		StrNCopy (paramP->str, str + paramP->start, len);
		paramP->str[len] = chrNull;
	}
	return true;
}


/***********************************************************************
 *
 * FUNCTION:    DASetFieldString
 *
 * DESCRIPTION: This routine sets string to the field.
 *
 * PARAMETERS:	paramP
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/18/04	Initial Revision
 *
 ***********************************************************************/
static void DASetFieldString (ParamType *paramP)
{
	UInt16		len = StrLen (paramP->str);

	FldReleaseFocus (paramP->fldP);
	//FldSetSelection (paramP->fldP, 0, 0);
	FldDelete (paramP->fldP, paramP->start, paramP->end);
	FldInsert (paramP->fldP, paramP->str, len);
	//FldSetSelection (paramP->fldP, paramP->end + len, paramP->end + len);
	//FldSetInsertionPoint (paramP->fldP, paramP->end + len);
	//FldGrabFocus (paramP->fldP);
}


/***********************************************************************
 *
 * FUNCTION:    InitForm
 *
 * DESCRIPTION: This routine sets items of application list 
 *				according to PsLink preferences and resize the form.
 *
 * PARAMETERS:	frmP - pointer to the form type structure
 *				paramP - Pointer to a ParamType structure
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/04/05	Initial Revision
 *
 ***********************************************************************/
static void InitForm (FormPtr frmP, ParamType *paramP)
{
	PsLinkPreferenceType	prefs;
	DmSearchStateType	searchState;
	UInt16				cardNo;
	LocalID				dbID;
	ListPtr				listP;
	RectangleType		r;
	WinHandle			winH;
	UInt16				i;
	UInt16				numApps = 0;
	UInt16				candi[] = {MEMO, TODO, ADDRESS, DATE, MORE1, MORE2, MORE3, MORE4, MORE5};


	// Get preferences of PsLink
	GetPreferences (&prefs);

	// Store application names to array
	for (i = 0; i < LENGTH(candi); i ++)
	{
		if (prefs.creator[candi[i]] != LINKAPP_NONE_ID &&
		 DmGetNextDatabaseByTypeCreator (true, &searchState,
		 sysFileTApplication, prefs.creator[candi[i]], false, &cardNo, &dbID) == errNone)
		{
			// Get application name
			if (i >= MEMO && i <= DATE)
			{ // from string list
				SysStringByIndex (AppNameStringList, candi[i],
				 paramP->appName[numApps], dmDBNameLength); 
			}
			else
			{ // from application resource
				DmDatabaseInfo (cardNo, dbID, paramP->appName[numApps], NULL,
				 NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
				GetApplName (cardNo, dbID, paramP->appName[numApps]);
			}
			paramP->array[numApps] = paramP->appName[numApps];
			paramP->appNum[numApps] = candi[i];
			numApps ++;
		}
	}

	// Set list item
	listP = (ListPtr)GetObjectPtr (SelectAppsList);
	LstSetListChoices (listP, paramP->array, numApps);
	LstSetHeight (listP, numApps);
	LstSetSelection (listP, 0);


	// Resize form
	winH = FrmGetWindowHandle (frmP);
	WinGetBounds (winH, &r);
	r.extent.y += numApps * 11;
	WinSetBounds (winH, &r);
}


/***********************************************************************
 *
 * FUNCTION:    Exit
 *
 * DESCRIPTION: This routine exits list.
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/04/05	Initial Revision
 *
 ***********************************************************************/
static void Exit (void)
{
	EventType		event;

	event.eType = appStopEvent;
	EvtAddEventToQueue (&event);
}


/***********************************************************************
 *
 * FUNCTION:    MoveSelection
 *
 * DESCRIPTION: This routine move a selection in the list item.
 *
 * PARAMETERS:	dir - direction to move
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/04/05	Initial Revision
 *
 ***********************************************************************/
static void MoveSelection (WinDirectionType dir)
{
	ListPtr		listP;
	UInt16		lastItem;
	Int16		select;


	listP = GetObjectPtr (SelectAppsList);
	lastItem = LstGetNumberOfItems (listP) - 1;
	select = LstGetSelection (listP);

	select += moveAmt(dir);
	if (select < 0)
		select = lastItem;
	else if (select > lastItem)
		select = 0;

	LstSetSelection (listP, select);
}


/***********************************************************************
 *
 * FUNCTION:    AppHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the 
 *              "SelectForm" of this application.
 *
 * PARAMETERS:  eventP - a pointer to an EventType structure
 *				paramP - a pointer to an ParamType structure
 *
 * RETURNED:    true if the event has handle and should not be passed
 *              to a higher level handler.
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/04/05	Initial Revision
 *
 ***********************************************************************/
static Boolean AppHandleEvent (EventPtr eventP, ParamType *paramP)
{
	RectangleType	r;
	Boolean			handled = false;


	switch (eventP->eType) 
	{
		case keyDownEvent:
			if (EvtKeydownIsVirtual(eventP))
			{
				if (NavDirectionPressed (eventP, Up)
				 || eventP->data.keyDown.chr == vchrThumbWheelUp
				 || eventP->data.keyDown.chr == vchrJogUp)
				{
					MoveSelection (winUp);
					handled = true;
				}
				else if (NavDirectionPressed (eventP, Down)
				 || eventP->data.keyDown.chr == vchrThumbWheelDown
				 || eventP->data.keyDown.chr == vchrJogDown)
				{
					MoveSelection (winDown);
					handled = true;
				}
				else if (NavDirectionPressed (eventP, Left)
				 || NavDirectionPressed (eventP, Right)
				 || eventP->data.keyDown.chr == vchrJogLeft
				 || eventP->data.keyDown.chr == vchrJogRight
				 || eventP->data.keyDown.chr == vchrJogBack)
				{
					paramP->appl = NONE;
					Exit ();
					handled = true;
				}
				else if (eventP->data.keyDown.chr == vchrJogPush)
				{
					paramP->jogPushed = true;
					handled = true;
				}
				else if (NavSelectPressed (eventP)
				 || eventP->data.keyDown.chr == vchrThumbWheelPush
				 || eventP->data.keyDown.chr == vchrJogRelease && paramP->jogPushed)
				{
					paramP->appl = paramP->appNum[LstGetSelection (GetObjectPtr (SelectAppsList))];
					Exit ();
					handled = true;
				}
			}
			break;

		case lstSelectEvent:
			paramP->appl = paramP->appNum[eventP->data.lstSelect.selection];
			Exit ();
			handled = true;
			break;					

		case penDownEvent:
			FrmGetFormBounds (FrmGetActiveForm (), &r);
			if (!RctPtInRectangle (eventP->screenX + r.topLeft.x, eventP->screenY + r.topLeft.y, &r))
			{
				paramP->appl = NONE;
				Exit ();
				handled = true;
			}
			break;

		default:
			break;
	}
	
	return handled;

}


/***********************************************************************
 *
 * FUNCTION:    AppEventLoop
 *
 * DESCRIPTION: This routine is the event loop for the application.  
 *
 * PARAMETERS:  paramP - a pointer to an ParamType structure
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/18/04	Initial Revision
 *
 ***********************************************************************/
static void AppEventLoop (ParamType *paramP)
{
	UInt16		error;
	EventType	event;

	do {
		EvtGetEvent(&event, evtWaitForever);

		if (! SysHandleEvent(&event))
			if (! MenuHandleEvent(0, &event, &error))
				if (! AppHandleEvent(&event, paramP))
					FrmDispatchEvent(&event);

	} while (event.eType != appStopEvent);
}


/***********************************************************************
 *
 * FUNCTION:	DAMain
 *
 * DESCRIPTION:	This is the main entry point for the Desc Accessory.
 *
 * PARAMETERS:	nothing
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/18/04	Initial Revision
 *
 ***********************************************************************/
void DAMain (void)
{
	FormPtr		frmP;
	MemHandle	paramH;
	ParamType	*paramP;
	UInt32		val;
	SonySysFtrSysInfoP	infoP = NULL;
	UInt16 		oldFtrCardNo, *ftrCardNoP;
	LocalID 	oldFtrDbID, *ftrDbIDP;
	UInt16	 	**jogAstMaskPP = NULL;
	UInt16 		*oldJogAstMaskP = NULL;
	UInt16 		jogAstMask[] = {sonyJogAstMaskType1, 1, SelectForm, sonyJogAstMaskAll};


	// Avoid double launching
	if (FtrGet (appFileCreator, FeatureDA, &val) == errNone)
		return;
	FtrSet (appFileCreator, FeatureDA, 0);

	// Allocate a chunk for parameter
	paramH = MemHandleNew (sizeof(ParamType));
	paramP = MemHandleLock (paramH);

	// Get selected string from the original field
	if (!DAGetFieldString (paramP))
		goto EXIT;

	// Mask jog assist
	if (FtrGet (sonySysFtrCreator, sonySysFtrNumSysInfoP, (UInt32 *)&infoP) == errNone
	 && infoP && infoP->extn & sonySysFtrSysInfoExtnJogAst
	 && FtrGet (sonySysFtrCreator, sonySysFtrNumJogAstMaskP, (UInt32*) &jogAstMaskPP) == errNone)
	{
		oldJogAstMaskP = *jogAstMaskPP;
		*jogAstMaskPP = jogAstMask;
		if (FtrGet (sonySysFtrCreator, sonySysFtrNumJogAstMOCardNoP, (UInt32*) &ftrCardNoP)== errNone
		 && FtrGet (sonySysFtrCreator, sonySysFtrNumJogAstMODbIDP, (UInt32*) &ftrDbIDP) == errNone)
		{
			oldFtrCardNo = *ftrCardNoP;
			oldFtrDbID = *ftrDbIDP;
			SysCurAppDatabase (ftrCardNoP, ftrDbIDP);
		}
	}

	// Initialize a main form
	frmP = FrmInitForm (SelectForm);
	FrmSetActiveForm (frmP);
	InitForm (frmP, paramP);
	paramP->appl = NONE;

	FrmDrawForm (frmP);

	// Event loop
	AppEventLoop(paramP);

	// Stop
	FrmEraseForm (frmP);
	FrmDeleteForm (frmP);

	// Restore jog assist state
	if (oldJogAstMaskP != NULL)
	{
		*jogAstMaskPP = oldJogAstMaskP;
		*ftrCardNoP = oldFtrCardNo;
		*ftrDbIDP = oldFtrDbID;
	}

	// Tagging
	if (paramP->appl != NONE)
	{
		GetDstString (paramP);
		DASetFieldString (paramP);
	}

	EXIT:
	// Dispose of a chunk
	MemHandleUnlock (paramH);
	MemHandleFree (paramH);

	FtrUnregister (appFileCreator, FeatureDA);
}

