/******************************************************************************
 *
 * Copyright (c) 2004 PalmSpace/Sembe.
 *				All rights reserved.
 *
 * File: PsLinkLib.c
 *
 * Description:
 *   This is the PsLink common library.
 *
 * History:
 *		Dec 18, 2004	Created by sembe
 *
 *****************************************************************************/

#include "PsLink.h"
#include "PsLinkLib.h"


/***********************************************************************
 *
 * FUNCTION:    GetObjectPtr
 *
 * DESCRIPTION: This routine returns a pointer to an object in the current
 *              form.
 *
 * PARAMETERS:  objectId - id of the object to get pointer
 *
 * RETURNED:    void *
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/29/03	Initial Revision
 *			smb		12/17/04	Return NULL if frmP is NULL or objectID is invalid.
 *
 ***********************************************************************/
void * GetObjectPtr (UInt16 objectID)
{
	FormPtr frmP;
	UInt16	index;

	if ((frmP = FrmGetActiveForm ()) != NULL
	 && (index = FrmGetObjectIndex (frmP, objectID)) != frmInvalidObjectId)
		return FrmGetObjectPtr (frmP, index);
	else
		return NULL;
}


/***********************************************************************
 *
 * FUNCTION:    FindStr
 *
 * DESCRIPTION: This routine finds string in string.
 *
 * PARAMETERS:  *s1
 *				*s2
 *				caseSensitive
 *				part
 *
 * RETURNED:	true if found
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/26/08	Initial Revision
 *
 ***********************************************************************/
Boolean FindStr (Char *s1, Char *s2, Boolean caseSensitive, Boolean part)
{
	Boolean	found;
	Char	targetString[0xff];
	UInt32	outPos;
	UInt16	outLength;


	if (part)
	{
		if (caseSensitive)
			found = (StrStr (s1, s2) != NULL);
		else
		{
			TxtGluePrepFindString (s2, targetString, 0xff);
			found = TxtFindString (s1, targetString, &outPos, &outLength);
		}
	}
	else
	{
		if (caseSensitive)
			found = (StrCompare (s1, s2) == 0);
		else
			found = (StrCaselessCompare (s1, s2) == 0);
	}
	return found;
}


/***********************************************************************
 *
 * FUNCTION:    GetApplName
 *
 * DESCRIPTION: This routine gets application name.
 *
 * PARAMETERS:  cardNo
 *				dbID
 *
 * RETURNED:    name
 *				true - if the application name exists
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/18/04	Initial Revision
 *
 ***********************************************************************/
Boolean GetApplName (UInt16 cardNo, LocalID dbID, Char *name)
{
	DmOpenRef	dbP;
	MemHandle	resH;
	Char		*resP;
	Boolean		exist = false;


	DmDatabaseInfo (cardNo, dbID, name, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
	dbP = DmOpenDatabase (cardNo, dbID, dmModeReadOnly);
	if (dbP != NULL)
	{
		resH = DmGet1Resource (ainRsc, ainID);
		if (resH)
		{
			resP = (Char *)MemHandleLock (resH);
			if (resP != NULL)
			{// Store application name instead of database name
				exist = true;
				StrCopy (name, resP);
				MemHandleUnlock (resH);
			}
			DmReleaseResource (resH);
		}
		DmCloseDatabase (dbP);
	}

	return exist;
}


/***********************************************************************
 *
 * FUNCTION:    SetPreferences
 *
 * DESCRIPTION: Save the current state of the application.
 *
 * PARAMETERS:  prefs
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/29/03	Initial Revision
 *
 ***********************************************************************/
void SetPreferences (PsLinkPreferenceType *prefs)
{

	PrefSetAppPreferences (appFileCreator, appPrefID, appPrefVersionNum, 
			prefs, sizeof (PsLinkPreferenceType), true);
}


/***********************************************************************
 *
 * FUNCTION:     GetPreferences
 *
 * DESCRIPTION:  Get the current application's preferences.
 *
 * PARAMETERS:   prefs
 *
 * RETURNED:     nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/29/03	Initial Revision
 *			smb		02/07/04	Modify to support ToDo and more other applications.
 *			smb		12/18/04	Support database link and database prefix
 *
 ***********************************************************************/
void GetPreferences (PsLinkPreferenceType *prefs)
{
	UInt16				prefsSize;
	Int16				prefsVer;
	UInt32				creatorID;
 	DmSearchStateType	state;
 	UInt16				cardNo;
	LocalID				dbID;
	UInt16				i;


	// Read the saved preferences / saved-state information.
	prefsSize = sizeof (PsLinkPreferenceType);
	prefsVer = PrefGetAppPreferences (appFileCreator, appPrefID, prefs, &prefsSize, true);
	if (prefsVer == appPrefVersionNum)
		return;

	if (prefsVer == noPreferenceFound)
	{// Set default value to the preferences, if no preferene found.
		prefs->state = IsOFF;
		prefs->singleTap = IsOFF;
		prefs->autoEnble = IsOFF;

		for (i = 0; i < NUM_LINK_APPS; i ++)
		{
			prefs->creator[i] = LINKAPP_NONE_ID;
			StrCopy (prefs->dbName[i], "");
			StrCopy (prefs->tag[i][START], "");
			StrCopy (prefs->tag[i][END], "");
		}

		if (ExgGetDefaultApplication (&creatorID, exgRegExtensionID, "txt") == errNone)
			prefs->creator[MEMO] = creatorID;
		else
			prefs->creator[MEMO] = sysFileCMemo;

		if (DmGetNextDatabaseByTypeCreator (true, &state, 'DATA', 'PTod', true, &cardNo, &dbID) == errNone)
			prefs->creator[TODO] = 'PTod';
		else
			prefs->creator[TODO] = sysFileCToDo;

		if (ExgGetDefaultApplication (&creatorID, exgRegExtensionID, "vcf") == errNone)
			prefs->creator[ADDRESS] = creatorID;
		else
			prefs->creator[ADDRESS] = sysFileCAddress;

		if (ExgGetDefaultApplication (&creatorID, exgRegExtensionID, "vcs") == errNone)
			prefs->creator[DATE] = creatorID;
		else
			prefs->creator[DATE] = sysFileCDatebook;

		if (ExgGetDefaultApplication (&creatorID, exgRegSchemeID, "http") == errNone)
			prefs->creator[WEB] = creatorID;
		else
			prefs->creator[WEB] = LINKAPP_NONE_ID;

		if (ExgGetDefaultApplication (&creatorID, exgRegSchemeID, "mailto") == errNone)
			prefs->creator[MAIL] = creatorID;
		else
			prefs->creator[MAIL] = LINKAPP_NONE_ID;

		StrCopy (prefs->tag[MEMO][START], TAG_MEMO_START);
		StrCopy (prefs->tag[MEMO][END], TAG_MEMO_END);
		StrCopy (prefs->tag[TODO][START], TAG_TODO_START);
		StrCopy (prefs->tag[TODO][END], TAG_TODO_END);
		StrCopy (prefs->tag[ADDRESS][START], TAG_ADDRESS_START);
		StrCopy (prefs->tag[ADDRESS][END], TAG_ADDRESS_END);
		for (i = NUM_DEF_LINK_APPS; i < NUM_LINK_APPS; i ++)
		{
			StrCopy (prefs->tag[i][START], TAG_MORE_START);
			StrPrintF (prefs->tag[i][END], TAG_MORE_END, i - NUM_DEF_LINK_APPS + 1);
		}
	}

	if (prefsVer <= 0x04)
	{
		prefs->caseSensitive = IsON;
		prefs->partOfTitle = IsOFF;

		if (prefsVer <= 0x03)
		{
			prefs->ignoreBtn = IsOFF;
			prefs->memoWikiName = IsOFF;
			prefs->todoCompleted = IsOFF;
			prefs->forwardKey.chr = 0;
			prefs->forwardKey.modifiers = 0;

			if (prefsVer <= 0x02)
			{
				prefs->backKey.chr = 0;
				prefs->backKey.modifiers = 0;

				if (prefsVer <= 0x01)
				{
					for (i = 0; i < NUM_LINK_APPS; i ++)
					{
						prefs->dbLink[i] = false;
						StrCopy (prefs->dbPrefix[i], DbPrefix(prefs->creator[i]));
					}

					if (prefsVer <= 0x00)
					{
						prefs->linkChrBookmark = LINKCHR_BOOKMARK;
						prefs->linkChrDBname = LINKCHR_DBNAME;
					}
				}
			}
		}
	}
	SetPreferences (prefs);
}


/***********************************************************************
 *
 * FUNCTION:    CheckDBOpenByCreator
 *
 * DESCRIPTION: This routine checks whether a specific database is opened.
 *
 * PARAMETERS:  dbP
 *				dbType
 *				dbCreator
 *				force - force to open db by dbID
 *
 * RETURNED:	true, if database is opened
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/03	Initial Revision
 *
 ***********************************************************************/
Boolean CheckDBOpenByCreator (DmOpenRef *dbP, UInt32 dbType, UInt32 dbCreator, Boolean force)
{
	UInt16		cardNo;
	LocalID		dbID;
	UInt32		type, creator;
	DmOpenRef	db;


	db = DmNextOpenDatabase (NULL);
	while (db)
	{
		DmOpenDatabaseInfo (db, &dbID, NULL, NULL, &cardNo, NULL);
		DmDatabaseInfo (cardNo, dbID, NULL, NULL, NULL, NULL,NULL,NULL,
						 NULL, NULL, NULL, &type, &creator);
		if (type == dbType && creator == dbCreator)
		{
			*dbP = db;
			return (true);
		}
		db = DmNextOpenDatabase (db);
	}

	if (force)
		*dbP = DmOpenDatabaseByTypeCreator (dbType, dbCreator, dmModeReadWrite);

	return (false);
}


/***********************************************************************
 *
 * FUNCTION:    CheckDBOpenByID
 *
 * DESCRIPTION: This routine checks whether a specific database is opened.
 *
 * PARAMETERS:  dbP
 *				mCardNo
 *				mdbID
 *				force - force to open db by dbID
 *
 * RETURNED:	true, if database is opened
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/03	Initial Revision
 *
 ***********************************************************************/
Boolean CheckDBOpenByID (DmOpenRef *dbP, UInt16 mCardNo, LocalID mdbID, Boolean force)
{
	DmOpenRef	db;
	UInt16		cardNo;
	LocalID		dbID;


	if (mdbID == NULL)
	{
		*dbP = NULL;
		return (false);
	}

	db = DmNextOpenDatabase (NULL);
	while (db)
	{
		DmOpenDatabaseInfo (db, &dbID, NULL, NULL, &cardNo, NULL);
		if (cardNo == mCardNo && dbID == mdbID)
		{
			*dbP = db;
			return (true);
		}
		db = DmNextOpenDatabase(db);
	}

	if (force)
		*dbP = DmOpenDatabase (mCardNo, mdbID, dmModeReadOnly);

	return (false);
}


/***********************************************************************
 *
 * FUNCTION:    CheckDBOpenByName
 *
 * DESCRIPTION: This routine checks whether a specific database is opened.
 *
 * PARAMETERS:  dbP
 *				dbName
 *				force - force to open db by dbID
 *
 * RETURNED:	true, if database is opened
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/09/03	Initial Revision
 *
 ***********************************************************************/
Boolean CheckDBOpenByName (DmOpenRef *dbP, Char *dbName, Boolean force)
{
	DmOpenRef	db;
	LocalID		dbID;
	UInt16		cardNo;
	Char		name[dmDBNameLength];


	db = DmNextOpenDatabase (NULL);
	while (db)
	{
		DmOpenDatabaseInfo (db, &dbID, NULL, NULL, &cardNo, NULL);
		DmDatabaseInfo (cardNo, dbID, name, NULL, NULL, NULL,NULL,NULL,
						 NULL, NULL, NULL, NULL, NULL);
		if (StrCompare(name, dbName) == 0)
		{
			*dbP = db;
			return (true);
		}
		db = DmNextOpenDatabase (db);
	}

	if (force)
	{
		*dbP = NULL;
		for (cardNo = 0; cardNo < MemNumCards (); cardNo ++)
		{
			dbID = DmFindDatabase (cardNo, dbName);
			if (dbID)
				*dbP = DmOpenDatabase (cardNo, dbID, dmModeReadOnly);
		}
	}

	return (false);
}


/***********************************************************************
 *
 * FUNCTION:    FindRecordByString
 *
 * DESCRIPTION: This routine finds a unique id of the record by string should be in.
 *
 * PARAMETERS:  dbP - pointer to the database
 *				string - string to match record
 *				recordNum - matched record
 *
 * RETURNED:	true if found
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/07/04	Initial Revision
 *			smb		12/17/04	Fix an error; Invalid char passed to TxtCharAttr
 *
 ***********************************************************************/
Boolean FindRecordByString (DmOpenRef dbP, Char *string, UInt16 *recordNum)
{
	UInt16		index;
	MemHandle	recordH;
	Char		*recordP;
	UInt32		recordSize;
	UInt16		offset;
	UInt16		len = StrLen (string);
	Boolean		found = false;


	*recordNum = NO_RECORD;

	index = dmMaxRecordIndex;
	while (!found)
	{
		if (DmSeekRecordInCategory (dbP, &index, 0, dmSeekBackward, dmAllCategories))
			break;
		recordH = DmQueryRecord (dbP, index);
		if (!recordH)
			break;
		recordSize = MemHandleSize (recordH);
		if (recordSize >= len)
		{
			recordP = (Char *)MemHandleLock (recordH);
			offset = 0;
			while (offset < recordSize && !found)
			{
				// Exclude non printable character
				while (offset < recordSize)
				{
					if (StrLen (recordP + offset) < len)
						offset ++;
					else
						break;
				}
				// Match string
				//if (StrNCompare (recordP + offset, string, len) == 0)
				if (offset >= recordSize)
					break;
				else if (StrStr (recordP + offset, string) != NULL)
				{
					*recordNum = index;
					found = true;
					break;
				}
				offset += StrLen (recordP + offset);
			}
			MemHandleUnlock (recordH);
		}
		if (index > 0)
			index --;
		else 
			break;
	}

	return (found);
}


/***********************************************************************
 *
 * FUNCTION:    GetBusyRecordID
 *
 * DESCRIPTION: This routine gets uniqueID of current record.
 *
 * PARAMETERS:  dbP - pointer to the database
 *				uniqueID - unique ID of current record
 *
 * RETURNED:	true if found
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		06/12/04	Initial Revision (separate from GetCurrentRecordID())
 *
 ***********************************************************************/
Boolean GetBusyRecordID (DmOpenRef dbP, UInt32 *uniqueIDP)
{
	UInt32		uniqueID;
	UInt16		index;
	UInt16		attr;


	*uniqueIDP = NO_UNIQUEID;

	if (dbP == NULL)
		return (false);

	// Try to find a record in use.
	index = 0;
	while (true)
	{
		if (DmSeekRecordInCategory (dbP, &index, 0, dmSeekForward, dmAllCategories))
			break;
		// look for a record under modification, where the 'busy' bit is set
		DmRecordInfo (dbP, index, &attr, NULL, NULL);
		if ((attr & dmRecAttrBusy) && !(attr & dmRecAttrDelete))
		{
			// get a unique ID from record number
			if (DmRecordInfo (dbP, index, NULL, &uniqueID, NULL) == errNone)
			{
				*uniqueIDP = uniqueID;
				return (true);
			}
			return (false);
		}
		index ++;
	}

	return (false);
}


/***********************************************************************
 *
 * FUNCTION:    SwitchApp
 *
 * DESCRIPTION: This routine switches application and go to specific record.
 *
 * PARAMETERS:	GoToParamsType parameters
 *
 * RETURNED:	error code
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/11/05	Initial Revision
 *
 ***********************************************************************/
Err SwitchApp (UInt16 appCardNo, LocalID appID,
 Int16 searchStrLen, UInt16 dbCardNo, LocalID dbID,
 UInt16 recordNum, UInt16 matchPos, UInt16 matchFieldNum, UInt32 matchCustom)
{
	Err					err;
	GoToParamsType		*cmdPBP;


	cmdPBP = (GoToParamsType *)MemPtrNew (sizeof (GoToParamsType));
	if (!cmdPBP)
		return (sysErrNoFreeRAM);

	cmdPBP->searchStrLen = searchStrLen;
	cmdPBP->dbCardNo = dbCardNo;
	cmdPBP->dbID = dbID;
	cmdPBP->recordNum = recordNum;
	cmdPBP->matchPos = matchPos;
	cmdPBP->matchFieldNum = matchFieldNum;
	cmdPBP->matchCustom = matchCustom;
	err = MemPtrSetOwner ((MemPtr)cmdPBP, 0);
	if (err == errNone)
		err = SysUIAppSwitch (appCardNo, appID, sysAppLaunchCmdGoTo, (MemPtr)cmdPBP);
	if (err)
		MemPtrFree (cmdPBP);

	return err;
}

