/******************************************************************************
 *
 * Copyright (c) 2004 PalmSpace/Sembe.
 *				All rights reserved.
 *
 * File: link.cpp
 *
 * Description:
 *   This is the PsLink application's sub module.
 *
 * History:
 *		Dec 29, 2003	Created by sembe
 *
 *****************************************************************************/

#include "PsLink.h"
#include "PsLinkLib.h"
#include "link.h"
#include "linkaction.h"
#include "PsLinkPlugin.h"
#include "AddressDB.h"
#include "DateDB.h"
#include "ToDoDB.h"


/***********************************************************************
 *
 * FUNCTION:    ExcludeSpaceInStr
 *
 * DESCRIPTION: This routine excludes space character from the string.
 *
 * PARAMETERS:	srcStr - source string
 *				dstStr - destination string
 *
 * RETURNED:	length of destination string
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/07/03	Initial Revision
 *
 ***********************************************************************/
static UInt16 ExcludeSpaceInStr (Char *srcStr, Char *dstStr)
{
	UInt16		len = 0;


	while (*srcStr)
	{
		if (*srcStr == chrSpace)
			srcStr ++;
		else if (StrChr (srcStr, chrSJISIdeographicSpace) == srcStr)
			srcStr += 2;
		else if (IsSJIS (*srcStr))
		{
			StrNCopy (dstStr + len, srcStr, 2);
			srcStr += 2;
			len += 2;
		}
		else
		{
			StrNCopy (dstStr + len, srcStr, 1);
			srcStr ++;
			len ++;
		}
	}
	dstStr[len] = chrNull;

	return (len);
}


/***********************************************************************
 *
 * FUNCTION:    LaunchWithPlugin
 *
 * DESCRIPTION: This routine launches application with plugin module.
 *
 * PARAMETERS:	creatorID
 *				dbName
 *				string - string to link
 *
 * RETURNED:	true if launched.
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/03	Initial Revision
 *			smb		02/21/04	Support plugins.
 *
 ***********************************************************************/
static Boolean LaunchWithPlugin (UInt32 creatorID, Char *dbName, Char *string)
{
	Err					err;
	Boolean				newSearch;
	UInt16				cardNo;
	LocalID				dbID;
	DmSearchStateType	state;
	UInt32				support, result;
	PluginParamsType	*cmdPBP;


	// Search plugin and launch it
	newSearch = true;
	while (true)
	{
		err = DmGetNextDatabaseByTypeCreator (newSearch, &state,
		 PsLinkTPlugins, NULL, false, &cardNo, &dbID);
		if (err)
			break;
		err = SysAppLaunch (cardNo, dbID, 0, PsLinkAppLaunchSupportCmd, NULL, &support);
		if (!err && support == creatorID)
		{
			cmdPBP = (PluginParamsType *)MemPtrNew (sizeof(PluginParamsType));
			if (!cmdPBP)
				return false;
			cmdPBP->creator = creatorID;
			StrCopy (cmdPBP->dbName, dbName);
			StrCopy (cmdPBP->string, string);
			MemPtrSetOwner (cmdPBP, 0);
			//SysUIAppSwitch (cardNo, dbID, PsLinkAppLaunchCmd, cmdPBP);
			err = SysAppLaunch (cardNo, dbID, 0, PsLinkAppLaunchCmd, cmdPBP, &result);
			MemPtrFree (cmdPBP);
			if (!err && result == SUCCESS)
				return true;
		}
		newSearch = false;
	}

	return false;
}


/***********************************************************************
 *
 * FUNCTION:    GoToURL
 *
 * DESCRIPTION: This routine open the browser application with URL.
 *
 * PARAMETERS:  browser - creator ID of the web browser
 *				origurl - ptr to URL string
 *
 * RETURNED:	err
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		01/05/03	Initial Revision
 *
 ***********************************************************************/
static Err GoToURL (UInt32 browser, Char *url)
{
	Err					err;
	Char				*cmdPBP;
	UInt16				cardNo;
	LocalID				dbID;
	DmSearchStateType	state;


	// find Viewer and launch it
	err = DmGetNextDatabaseByTypeCreator (true, &state,
		 sysFileTApplication, browser, true, &cardNo, &dbID);
	if (!err)
	{
		// make a copy of the URL, since the OS will free
		// the parameter once Viewer quits
		cmdPBP = (Char *)MemPtrNew (StrLen (url) + 1);
		if (!cmdPBP)
			return sysErrNoFreeRAM;

		StrCopy (cmdPBP, url);
		MemPtrSetOwner (cmdPBP, 0);
		err = SysUIAppSwitch (cardNo, dbID, sysAppLaunchCmdGoToURL, cmdPBP);
		if (err)
			MemPtrFree (cmdPBP);
	}

	return err;
}


/***********************************************************************
 *
 * FUNCTION:    SendMail
 *
 * DESCRIPTION: This routine open the Mail apprication.
 *
 * PARAMETERS:  creator - creator ID of the mail
 *				mailTo - ptr to mailto string
 *
 * RETURNED:	err
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/21/04	Initial Revision
 *
 ***********************************************************************/
static Err SendMail (UInt32 creator, Char *mailTo)
{
	Err					err;
	UInt16				cardNo;
	LocalID				dbID;
	DmSearchStateType	state;
	Char				*cmdPBP;


	// find Viewer and launch it
	err = DmGetNextDatabaseByTypeCreator (true, &state,
		 sysFileTApplication, creator, true, &cardNo, &dbID);

	if (!err)
	{
		cmdPBP = (Char *)MemPtrNew (StrLen ("mailto:") + StrLen (mailTo) + 1);
		if (!cmdPBP)
			return sysErrNoFreeRAM;

		StrCopy (cmdPBP, "mailto:");
		StrCat (cmdPBP, mailTo);

		MemPtrSetOwner(cmdPBP, 0);
		err = SysUIAppSwitch (cardNo, dbID, sysAppLaunchCmdGoToURL, cmdPBP);
		if (err)
			MemPtrFree (cmdPBP);
	}

	return err;

}


/***********************************************************************
 *
 * FUNCTION:    LaunchWeb
 *
 * DESCRIPTION: This routine launch web browser.
 *
 * PARAMETERS:  string
 *				creator
 *
 * RETURNED:	true if launched.
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/03	Initial Revision
 *			smb		02/21/04	Support plugins.
 *
 ***********************************************************************/
static Boolean LaunchWeb (UInt32 creator, Char *string)
{

	if (LaunchWithPlugin (creator, "", string)
	 || GoToURL (creator, string) == errNone)
	{// Save db information to be linked
		SaveLinkedDBInfo (creator, NULL, NULL, NULL, NULL, NO_RECORD);
		return true;
	}

	return false;
}


/***********************************************************************
 *
 * FUNCTION:    LaunchMail
 *
 * DESCRIPTION: This routine launch mailer.
 *
 * PARAMETERS:	creatorID
 *				mailTo
 *
 * RETURNED:	true if launched.
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/03	Initial Revision
 *			smb		02/21/04	Support plugins.
 *
 ***********************************************************************/
static Boolean LaunchMail (UInt32 creatorID, Char *mailTo)
{
	if (!LaunchWithPlugin (creatorID, "", mailTo))
		SendMail (creatorID, mailTo);

	return true;
}


/***********************************************************************
 *
 * FUNCTION:    LaunchMemo
 *
 * DESCRIPTION: This routine launches memo application.
 *
 * PARAMETERS:	creatorID
 *				string
 *				chrBookmark
 *
 * RETURNED:	true if launched.
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/03	Initial Revision
 *			smb		02/12/04	Modify the way to match string. 
 *			smb		06/01/04	Modify to deal with "MemosDB-PMem" 
 *								and support bookmark-link as a feature.
 *			smb		12/10/04	Support bookmark-link in the memo itself.
 *			smb		12/26/08	Support "case sensitive" and "use part of title".
 *
 ***********************************************************************/
static Boolean LaunchMemo (UInt32 creatorID, Char *str, PsLinkPreferenceType *prefs)
{
	Err				err;
	DmOpenRef		dbP;
	Boolean			open, part;
	MemHandle		recordH;
	Char			*recordP;
	Char			title[MAXLEN_LINKSTRING + 1];
	UInt16			titleLen, len;
	UInt16			recordNum, index;
	UInt16			matchPos = 0;
	UInt32			matchLen = 0;
	UInt32			uniqueID;
	Boolean			match;
	DmSearchStateType	state;
	LocalID			appID, dbID;
	UInt16			appCardNo, dbCardNo;
	Char			*matchStr, *ptr;


	// check all the open database and see if memo is currently open
	open = CheckDBOpenByCreator (&dbP, 'DATA', sysFileCMemo, true);
	if (dbP == NULL)
	{
		open = CheckDBOpenByCreator (&dbP, 'DATA', 'PMem', true);
		if (dbP == NULL)
			return false;
	}

	// get memo title and search string from str
	if (prefs->linkChrBookmark != chrNull)
	{
		matchStr = StrChr (str, prefs->linkChrBookmark);
		if (StrLen (str) == (UInt16)(matchStr - str) + 1)
			matchStr = NULL;
	}
	else
		matchStr = NULL;

	if (matchStr != NULL && matchStr == str)
	{
		matchStr ++;
		if (open)
		{
			if (GetBusyRecordID (dbP, &uniqueID)
			 && DmFindRecordByID (dbP, uniqueID, &recordNum) == errNone)
				match = true;
			else
				return false;
		}
		else
		{
			DmCloseDatabase (dbP);
			return false;
		}
	}
	else
	{
		if (matchStr != NULL)
			StrNCopy (matchStr ++, "\0", 1);

		// Search memo record which has specific title
		match = false;
		index = 0;
		recordNum = NO_RECORD;
		part = prefs->partOfTitle;
		while (!match)
		{
			recordH = DmQueryNextInCategory (dbP, &index, dmAllCategories);
			if (recordH == NULL)
				break;
			recordP = (Char *)MemHandleLock (recordH);
			ptr = StrChr (recordP, linefeedChr);
			titleLen = (ptr == NULL)? StrLen (recordP) : (UInt16)(ptr - recordP);
			StrNCopy (title, recordP, titleLen);
			title[titleLen] = chrNull;
			if (FindStr (title, str, prefs->caseSensitive, part))
			{
				recordNum = index;
				if (part)
				{
					part = false;
					index --;
				}
				else
					match = true;
			}
			MemHandleUnlock (recordH);
			index ++;
		}
		if (recordNum != NO_RECORD)
			match = true;
	}

	if (match)
	{
		recordH = DmQueryRecord (dbP, recordNum);
		recordP = (Char *)MemHandleLock (recordH);
		if (matchStr != NULL)
		{// Search string in the memo
			ptr = recordP;
			while ((ptr = StrStr (ptr, matchStr)) != NULL)
			{
				if (*(ptr - 1) != prefs->linkChrBookmark)
				{
					matchPos = (UInt16)(ptr - recordP);
					matchLen = StrLen (matchStr);
					break;
				}
				ptr += StrLen (matchStr);
			}
		}
		MemHandleUnlock (recordH);
	}
	else
	{// Create new record, if record is not found.
		len = StrLen (str);
		recordNum = dmMaxRecordIndex;
		recordH = DmNewRecord (dbP, &recordNum, len + 1);
		if (recordH)
		{
			recordP = (Char *)MemHandleLock (recordH);
			DmSet (recordP, 0, len + 1, 0);
			err = DmWrite (recordP, 0, str, len);
			ErrFatalDisplayIf (err,"DmWrite");
			DmReleaseRecord (dbP, recordNum, true);
			MemHandleUnlock (recordH);
		}
	}

	DmOpenDatabaseInfo (dbP, &dbID, NULL, NULL, &dbCardNo, NULL);
	if (!open)
	{
		DmCloseDatabase (dbP);
	}

	if (DmGetNextDatabaseByTypeCreator (true, &state,
	 sysFileTApplication, creatorID, true, &appCardNo, &appID) == errNone
	 && SwitchApp (appCardNo, appID, 0, dbCardNo, dbID, recordNum, matchPos, 0, matchLen) == errNone)
	{ // Save db information to be linked
		SaveLinkedDBInfo (creatorID, appCardNo, appID, dbCardNo, dbID, recordNum);
		return true;
	}

	return false;
}


/***********************************************************************
 *
 * FUNCTION:    LaunchToDo
 *
 * DESCRIPTION: This routine launches ToDo application.
 *
 * PARAMETERS:	creatorID
 *				string
 *
 * RETURNED:	true if launched.
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/07/04	Initial Revision
 *			smb		02/12/04	Modify the way to match string. 
 *			smb		01/29/05	Deal with "TasksDB-PTod" on the PalmOne devices 
 *			smb		03/26/05	Support "link to completed items" option.
 *			smb		12/26/08	Support "case sensitive" and "use part of title".
 *
 ***********************************************************************/
static Boolean LaunchToDo (UInt32 creatorID, Char *string, PsLinkPreferenceType *prefs)
{
	Err					err;
	DmOpenRef			dbP;
	Boolean				open, part;
	Int32				len;
	UInt16				index, recordNum;
	MemHandle			recH;
	ToDoDBRecordPtr		tRecP;
	TasksDBRecordPtr	pRecP;
	ToDoItemType		newToDo;
	Boolean				match = false;
	DmSearchStateType	state;
	UInt16				dbCardNo, appCardNo;
	LocalID				dbID, appID;
	Boolean				isPTod = false;


	open = CheckDBOpenByCreator (&dbP, 'DATA', sysFileCToDo, true);
	if (dbP == NULL)
	{ // try to open "Tasks" databae
		open = CheckDBOpenByCreator (&dbP, 'DATA', 'PTod', true);		
		if (dbP == NULL)
			return false;
		isPTod = true;
	}

	len = StrLen (string);
	// Try to match string completely
	index = 0;
	recordNum = NO_RECORD;
	part = prefs->partOfTitle;
	while (!match)
	{
		if ((recH = DmQueryNextInCategory (dbP, &index, dmAllCategories)) == NULL)
			break;
		if (isPTod	// Tasks
		 && (pRecP = (TasksDBRecordPtr)MemHandleLock (recH)) != NULL
		 && (prefs->todoCompleted || !(pRecP->flag & 0x40))	// not completed?
		 && pRecP->flag & 0x08	// has title?
		 && FindStr (&pRecP->description + ((pRecP->flag & 0xc0)? sizeof(DateType) : 0), string, prefs->caseSensitive, part)
		|| !isPTod	// ToDo
		 && (tRecP = (ToDoDBRecordPtr)MemHandleLock (recH)) != NULL
		 && (prefs->todoCompleted || !(tRecP->priority & completeFlag))
		 && tRecP->description != NULL
		 && FindStr (&tRecP->description, string, prefs->caseSensitive, part))
		{
			recordNum = index;
			if (part)
			{
				part = false;
				index --;
			}
			else
				match = true;
		}
		MemHandleUnlock (recH);
		index ++;
	}
	if (recordNum != NO_RECORD)
		match = true;

	if (!match)
	{ // Create a new record, if no record is matched.
		newToDo.priority = defaultPriority;
		*((UInt16 *) &newToDo.dueDate) = toDoNoDueDate;
		newToDo.description = string;
		newToDo.note = NULL;

		if (isPTod)
			err = TasksNewRecord (dbP, &newToDo, dmUnfiledCategory, &recordNum);
		else
			err = ToDoNewRecord (dbP, &newToDo, dmUnfiledCategory, &recordNum);
		if (err)
		{
			FrmAlert(DeviceFullAlert);
			return false;
		}
	}

	DmOpenDatabaseInfo (dbP, &dbID, NULL, NULL, &dbCardNo, NULL);
	if (!open)
	{ // Close the database, if it had not been opened.
		DmCloseDatabase (dbP);
	}

	if (DmGetNextDatabaseByTypeCreator (true, &state,
	 sysFileTApplication, creatorID, true, &appCardNo, &appID) == errNone
	 && SwitchApp (appCardNo, appID, 0, dbCardNo, dbID, recordNum, 0, 0, 0) == errNone)
	{ // Save db information to be linked
		SaveLinkedDBInfo (creatorID, appCardNo, appID, dbCardNo, dbID, recordNum);
		return true;
	}

	return false;
}


/***********************************************************************
 *
 * FUNCTION:    LaunchAddress
 *
 * DESCRIPTION: This routine launches address application.
 *
 * PARAMETERS:	creatorID
 *				string
 *
 * RETURNED:	true if launched.
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/03	Initial Revision
 *			smb		02/07/04	Fix bug; Error if the field in the address record is null.
 *			smb		03/26/05	Fix bug; Can't link from Contacts to Contacts.
 *			smb		08/08/05	Fix bug for AgendusPro.
 *
 ***********************************************************************/
static Boolean LaunchAddress (UInt32 creatorID, Char *string)
{
	Err					err;
	DmOpenRef			dbP;
	Boolean				open, openPdb = false;
	Int16				i;
	UInt16				recordNum;
	MemHandle			recordH;
	AddrDBRecordType	record;
	Boolean				match;
	UInt16				len;
	UInt16				matches1, matches2;
	UInt16				fieldNum[] = {company, name, firstName, name, company};
	Char				matchString[MAXLEN_LINKSTRING + 1];
	DmSearchStateType	state;
	UInt16				dbCardNo, appCardNo;
	LocalID				dbID, appID;


	// Close ContactsDB-PAd if it opens.
	if (CheckDBOpenByCreator (&dbP, 'DATA', 'PAdd', false))
	{
		DmCloseDatabase (dbP);
		openPdb = true;
	}

	// Open AddressDB
	open = CheckDBOpenByCreator (&dbP, 'DATA', sysFileCAddress, true);
	if (dbP == NULL)
		return false;

	// Initialize parameters
	match = false;
	len = StrLen (string);

	// Try to match string completely
	recordNum = 0;
	while (true)
	{
		recordH = DmQueryNextInCategory (dbP, &recordNum, dmAllCategories);
		if (!recordH)
			break;

		// Search all the fields of the address record.
		AddrDBGetRecord (dbP, recordNum, &record, &recordH);
		for (i = name; i <= company; i ++)
		{
			if (record.fields[i] != NULL
			 && StrStr (record.fields[i], string) != NULL)
			{
				match = true;
				break;
			}
		}
		MemHandleUnlock (recordH);
		if (match)
			break;
		recordNum ++;
	}

	if (!match)
	{// Try to match both lastname and firstname
		// Exclude space character from string to match.
		len = ExcludeSpaceInStr (string, matchString);

		recordNum = 0;
		while (true)
		{
			recordH = DmQueryNextInCategory (dbP, &recordNum, dmAllCategories);
			if (!recordH)
				break;

			// PrvAppSearch all the fields of the address record.
			AddrDBGetRecord (dbP, recordNum, &record, &recordH);

			matches1 = 0;
			for (i = 0; i < LENGTH(fieldNum); i ++)
			{
				if (record.fields[fieldNum[i]] != NULL)
				{
					TxtCaselessCompare (matchString + matches1, 0xFFFF, &matches2,
									 record.fields[fieldNum[i]], 0xFFFF, NULL);
					matches1 += matches2;
					if (matches1 == len)
					{
						match = true;
						break;
					}
				}
			}

			MemHandleUnlock (recordH);
			if (match)
				break;
			recordNum ++;
		}
	}

	if (!match)
	{// Create a new record, if no record is matched.
		for (i = firstAddressField; i < addressFieldsCount; i++)
		{
			record.fields[i] = NULL;
		}
		record.fields[name] = string;

		err = AddrDBNewRecord(dbP, &record, &recordNum);
		if (err)
		{
			FrmAlert(DeviceFullAlert);
			return false;
		}
	}

	DmOpenDatabaseInfo (dbP, &dbID, NULL, NULL, &dbCardNo, NULL);

	// Close AddressDB if it was not opened.
	if (!open)
	{
		DmCloseDatabase (dbP);
	}

	// Re-open ContactsDB-PAd if the database was opened.
	if (openPdb
	 && DmOpenDatabaseByTypeCreator ('DATA', 'PAdd', dmModeReadWrite) == NULL)
	 	return false;

	// Launch the application
	if (DmGetNextDatabaseByTypeCreator (true, &state,
	 sysFileTApplication, creatorID, true, &appCardNo, &appID) == errNone
	 && SwitchApp (appCardNo, appID, 0, dbCardNo, dbID, recordNum, 0, 0, 0) == errNone)
	{ // Save db information to be linked
		SaveLinkedDBInfo (creatorID, appCardNo, appID, dbCardNo, dbID, recordNum);
		return true;
	}

	return false;
}


/***********************************************************************
 *
 * FUNCTION:    LaunchDate
 *
 * DESCRIPTION: This routine launches date application.
 *
 * PARAMETERS:	creator - id of creator to launch
 *				date - date type
 *
 * RETURNED:	true if launched.
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/03	Initial Revision
 *			smb		02/05/04	Modify to return if no record in date db.
 *			smb		01/29/05	Deal with "Calendar" on the PalmOne devices.
 *			smb		02/13/05	Jump to the day evenif no schedule found in the day.
 *			smb		03/26/05	Modify not to use CalendarDB-PDa.
 *			smb		08/08/05	Fix bug for AgendusPro.
 *
 ***********************************************************************/
static Boolean LaunchDate (UInt32 creatorID, DateType jumpDate)
{
	Err					err;
	DmOpenRef			dbP;
	MemHandle			recordH;
	ApptPackedDBRecordPtr	recordP;
	UInt16				index, recordNum;
	Boolean				open, openPdb = false;
	Boolean				match;
	DateType			date;
	UInt32				days, jumpDays;
	LocalID				appID, dbID;
	UInt16				appCardNo, dbCardNo;
	DmSearchStateType	state;


	// Close CalendarDB-PDa if it opens.
	if (CheckDBOpenByCreator (&dbP, 'DATA', 'PDat', false))
	{
		DmCloseDatabase (dbP);
		openPdb = true;
	}

	// check all the open database and see if memo is currently open
	open = CheckDBOpenByCreator (&dbP, 'DATA', sysFileCDatebook, true);
	if (dbP == NULL)
		return false;

	jumpDays = DateToDays (jumpDate);

	match = false;
	index = 0;
	while (true)
	{
		recordH = DmQueryNextInCategory (dbP, &index, dmAllCategories);
		if (!recordH)
			break;

		recordP = (ApptPackedDBRecordPtr)MemHandleLock (recordH);
		date = recordP->when.date;
		MemHandleUnlock (recordH);

		days = DateToDays (date);

		if (days == jumpDays)
		{
			recordNum = index;
			match = true;
			break;
		}
		index ++;
	}

	if (!match)
	{// create a new record with no time specification
		TimeType	time;

		time.hours = apptNoTime;
		time.minutes = apptNoTime;
		err = CreateNewAppt (dbP, jumpDate, time, time, " ", &recordNum);
		if (!err)
		{
			match = true;
			EvtEnqueueKey (chrGroupSeparator, 0, 0);
			EvtEnqueueKey (chrBackspace, 0, 0);
		}
	}

	DmOpenDatabaseInfo (dbP, &dbID, NULL, NULL, &dbCardNo ,NULL);
	if (!open)
		DmCloseDatabase (dbP);

	if (!match)
		return false;

	// Re-open CalendarDB-PDa if the database was opened.
	if (openPdb
	 && DmOpenDatabaseByTypeCreator ('DATA', 'PDat', dmModeReadWrite) == NULL)
	 	return false;

	if (DmGetNextDatabaseByTypeCreator (true, &state,
	 sysFileTApplication, creatorID, true, &appCardNo, &appID) == errNone
	 && SwitchApp (appCardNo, appID, 0, dbCardNo, dbID, recordNum, 0, 0, 0) == errNone)
	{ // Save db information to be linked
		SaveLinkedDBInfo (creatorID, appCardNo, appID, dbCardNo, dbID, recordNum);
		return true;
	}

	return false;
}


/***********************************************************************
 *
 * FUNCTION:    LaunchApp
 *
 * DESCRIPTION: This routine launches application.
 *
 * PARAMETERS:	creatorID
 *				string
 *
 * RETURNED:	true if launched.
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/08/04	Initial Revision
 *			smb		02/23/04	Support plugins.
 *			smb		02/29/04	Support database name link.
 *			smb		12/18/04	Support DB name prefix and bookmark.
 *			smb		02/18/07	Address list items of Shadow Plan.
 *
 ***********************************************************************/
static Boolean LaunchApp (UInt32 creatorID, Char *dbName, Char *string,
 Boolean dbLink, Char *dbPrefix, Char chrDBname, Char chrBookmark)
{
	Err					err;
	DmOpenRef			dbP;
	Boolean				open;
	Boolean				newSearch;
	UInt16				recordNum;
	Boolean				match = false;
	DmSearchStateType	state;
	UInt16				dbCardNo, appCardNo;
	LocalID				dbID, appID;
	UInt32				type;
	UInt16				attr;
	Char				linkDbName[dmDBNameLength];
	Char				*matchStr;
	UInt16				matchPos, matchFieldNum;
	UInt32				matchCustom;


	// Launch application with plugin, if plugin module is installed.
	if (LaunchWithPlugin (creatorID, dbName, string))
	{
		SaveLinkedDBInfo (creatorID, NULL, NULL, NULL, NULL, NO_RECORD);
		return true;
	}

	// Find a record which the specific string is in.
	if (string[0] == chrDBname || dbLink)
	{// Link database which name is 'string'
		if (chrBookmark != chrNull)
		{
			matchStr = StrChr (string, chrBookmark);
			if (matchStr == NULL || StrLen (string) == (UInt16)(matchStr - string) + 1)
				matchStr = NULL;
			else
				StrCopy (matchStr ++, "\0");
		}
		else
			matchStr = NULL;

		if (string[0] == chrDBname)
			StrCopy (linkDbName, string + 1);
		else
			StrPrintF (linkDbName, "%s%s", dbPrefix, string);

		for (dbCardNo = 0; dbCardNo < MemNumCards (); dbCardNo ++)
		{
			dbID = DmFindDatabase (dbCardNo, linkDbName);
			if (dbID)
			{
				match = true;
				recordNum = NO_RECORD;
				if (matchStr != NULL)
				{
					open = CheckDBOpenByID (&dbP, dbCardNo, dbID, true);
					if (dbP)
					{
						match = FindRecordByString (dbP, matchStr, &recordNum);
						if (!open)
							DmCloseDatabase (dbP);
					}
				}
				break;
			}
		}
	}
	else if (*dbName == NULL)
	{// if dbName is "Default", search all database which creator is 'creatorID'.
		newSearch = true;
		while (!match)
		{
			err = DmGetNextDatabaseByTypeCreator (newSearch, &state,
				 NULL, DbCreator(creatorID), false, &dbCardNo, &dbID);
			if (err)
				break;
			DmDatabaseInfo (dbCardNo, dbID, NULL, &attr, NULL, NULL,
				 NULL, NULL, NULL, NULL, NULL, &type, NULL);
			if (IsData(type) && !(attr & dmHdrAttrResDB))
			{
				open = CheckDBOpenByID (&dbP, dbCardNo, dbID, true);
				if (dbP)
				{
					match = FindRecordByString (dbP, string, &recordNum);
					if (!open)
						DmCloseDatabase (dbP);
				}
			}
			newSearch = false;
		}
	}
	else
	{// Search database which name is 'dbName'.
		open = CheckDBOpenByName (&dbP, dbName, true);
		if (dbP == NULL)
			return false;
		match = FindRecordByString (dbP, string, &recordNum);
		DmOpenDatabaseInfo (dbP, &dbID, NULL, NULL, &dbCardNo, NULL);
		if (!open)
			DmCloseDatabase (dbP);
	}

	DmGetNextDatabaseByTypeCreator (true, &state,
	 sysFileTApplication, creatorID, true, &appCardNo, &appID);

	if (match)
	{// Launch application and goto matched record.
		// Parameters
		matchPos = 0;
		matchFieldNum = 0;
		matchCustom = 0;
		if (creatorID == 'NOtl')
			matchCustom = 0x0000044c;
		else if (creatorID == 'Coog')
		{
			matchFieldNum = recordNum - 1;
			recordNum = 0;
		}
		// Launch!
		err = SwitchApp (appCardNo, appID, 0, dbCardNo, dbID, recordNum, matchPos, matchFieldNum, matchCustom);
		//err = SwitchApp (appCardNo, appID, 0, dbCardNo, dbID, recordNum, 0, 0, 0x030000000);
	}

	if (!match || err)
	{// Add string to cliipboard and launch application, if no record is matched.
		ClipboardAddItem (clipboardText, string, StrLen (string));
		err = SysUIAppSwitch (appCardNo, appID, sysAppLaunchCmdNormalLaunch, NULL);
	}

	if (err == errNone)
	{ // Save db information to be linked
		SaveLinkedDBInfo (creatorID, appCardNo, appID, dbCardNo, dbID, recordNum);
		return true;
	}

	return false;
}


/***********************************************************************
 *
 * FUNCTION:    MatchMailTo
 *
 * DESCRIPTION: This routine matches mailto pattern.
 *
 * PARAMETERS:  fieldtext - pointer to the string
 *				fieldsize - string size
 *				pos - start position to match
 *				startP - start position of URL
 *				endP - end position of URL
 *
 * RETURNED:	true - match mailto
 *				false - no mailto
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		01/19/03	Initial Revision
 *
 ***********************************************************************/
static Boolean MatchMailTo (Char *fieldtext, UInt16 fieldsize, UInt16 pos,
 UInt16 *startP, UInt16 *endP)
{
	UInt16		start, end;
	UInt16 		find;
	UInt16 		i;


	if (pos < 1
	 || !IsURL(fieldtext[pos - 1]) || !IsURL(fieldtext[pos]))
		return (false);

	// Find the start point
	for (start = pos - 1; start > 0; start --)
	{
		if (!IsURL(fieldtext[start - 1]))
			break;
	}

	// Find the end point
	for (end = pos + 1; end < fieldsize; end ++)
	{
		if (!IsURL(fieldtext[end]))
			break;
	}

	// include mailto pattern?
	find = 0;
	for (i = start; i < end; i ++)
	{
		if (find == 0 && fieldtext[i] == '@')
			find = 1;
		else if (find == 1 && fieldtext[i] == '.')
		{
			find = 2;
			break;
		}
	}
	if (i >= end || find < 2)
		return (false);

	*startP = start;
	*endP = end;

	return (true);

}


/***********************************************************************
 *
 * FUNCTION:    MatchURL
 *
 * DESCRIPTION: This routine matches URL pattern.
 *
 * PARAMETERS:  fieldtext - pointer to the string
 *				fieldsize - string size
 *				pos - start position to match
 *				urlTag - tag which indicate start position
 *				startP - start position of URL
 *				endP - end position of URL
 *
 * RETURNED:	true - match URL
 *				false - no URL
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/02	Initial Revision
 *
 ***********************************************************************/
static Boolean MatchURL (Char *fieldtext, UInt16 fieldsize, UInt16 pos,
 Char *urlTag, UInt16 *startP, UInt16 *endP)
{
	UInt16		urlTagLen;
	UInt16		start;
	UInt16		end;


	if (pos < 1)
		return (false);

	// Set the string to find.
	urlTagLen = StrLen (urlTag);

	// Find the start code of URL
	start = pos - 1;
	while (true)
	{
		if (StrNCompare (fieldtext + start, urlTag, urlTagLen) == 0)
			break;
		else if (!IsURL(fieldtext[start]) || start <= 0)
			return (false);
		else
			start --;
	}

	// Find the end of URL
	for (end = pos + 1; end < fieldsize; end ++)
	{
		if (!IsURL(fieldtext[end]))
			break;
	}

	if (start == end)
		return (false);

	*startP = start;
	*endP = end;

	return (true);

}


/***********************************************************************
 *
 * FUNCTION:    MatchTag
 *
 * DESCRIPTION: This routine matches URL pattern.
 *
 * PARAMETERS:  fieldtext - pointer to the string
 *				fieldsize - string size
 *				pos - start position to match
 *				startTag - tag which indicate start position
 *				endTag - tag which indicate end position
 *				startP - start position of URL
 *				endP - end position of URL
 *
 * RETURNED:	true - match URL
 *				false - no URL
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/02	Initial Revision
 *			smb		02/17/04	Modify to fix bug; link text should not include tag.
 *
 ***********************************************************************/
static Boolean MatchTag (Char *string, UInt16 size, UInt16 pos,
 Char *startTag, Char *endTag, UInt16 *startP, UInt16 *endP)
{
	UInt16		startTagLen, endTagLen;
	UInt16		start, matchStart, matchEnd;
	UInt16		end;
	Boolean		match;


	// Set the string to find.
	startTagLen = StrLen (startTag);
	endTagLen = StrLen (endTag);

	if (pos < startTagLen)
		return (false);
	
	// Find the start tag
	start = matchEnd = 0;
	match = false;
	while (start <= pos - startTagLen)
	{
		if (StrNCompare (string + start, startTag, startTagLen) == 0)
		{
			matchStart = start;
			match = true;
		}
		else if (StrNCompare (string + start, endTag, endTagLen) == 0)
			matchEnd = start;
		if (IsSJIS(string[start]))
			start += 2;
		else
			start ++;
	}
	if (!match || matchStart < matchEnd)
		return (false);
	matchStart += startTagLen;

	// Find the end tag
	end = pos;
	while (true)
	{
		if (StrNCompare (string + end, endTag, endTagLen) == 0)
			break;
		else if (StrNCompare (string + start, startTag, startTagLen) == 0
		 || end + endTagLen > size)
			return (false);
		else
		{
			if (IsSJIS(string[end]))
				end += 2;
			else
				end ++;
		}
	}

	if (matchStart == end)
		return (false);

	*startP = matchStart;
	*endP = end;

	return (true);

}


/***********************************************************************
 *
 * FUNCTION:    MatchWikiName
 *
 * DESCRIPTION: This routine matches WikiName pattern.
 *
 * PARAMETERS:	buf - pointer to the string
 *				start - start position of WikiName
 *				end - end position of WikiName
 *
 * RETURNED:	true - match
 *				false - not match
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		03/26/05	Initial Revision (from PalmWiki)
 *
 ***********************************************************************/
static Boolean MatchWikiName (Char *buf, UInt16 pos, UInt16 *start, UInt16 *end)
{
	int wstart, wend;
	int state;
	int i;
	char c;
	int l;
	int buflen = StrLen(buf);
	//
	// Check WikiNames
	//
	*start = *end = pos;
	wstart = l = 0;
	while (l < *start) {
		if (IsSJIS(buf[l])) {
			l += 2;
			wstart = l;
			continue;
		}
		if (!IsAlpha(buf[l])) {
			l++;
			wstart = l;
			continue;
		}
		l++;
	}
	for (wend = *end; wend < buflen; wend ++) {
		if (!IsAlpha(buf[wend])) break;
	}

	// check if buf[wstart..wend] is an WikiName
	state = 0;
	for (i = wstart; i < wend; i ++) {
		c = buf[i];
		switch(state){
		case 0:	if (IsUpper(c)) state = 1;
			else state = -1;
			break;
		case 1:	if (IsLower(c)) state = 2;
			else state = -1;
			break;
		case 2:	if (IsUpper(c)) state = 3;
			else if (IsLower(c)) state = 2;
			else state = -1;
			break;
		case 3:	if (IsLower(c)) state = 4;
			else state = -1;
			break;
		case 4:	if (IsUpper(c)) state = 3;
			else if (IsLower(c)) state = 4;
			else state = -1;
			break;
		default:
			break;
		}
	}
	if (state == 4) {
		*start = wstart;
		*end = wend;
		return true;
	}
	return false;
}


/***********************************************************************
 *
 * FUNCTION:    Get3Val
 *
 * DESCRIPTION: This routine gets 3 value as date type.
 *
 * PARAMETERS:
 *
 * RETURNED:	true - if match
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/02	Initial Revision (from PalmWiki)
 *
 ***********************************************************************/
static Boolean Get3Val (Char *s, Int16 *ss, Int16 *ee,
	 Int16 *val1, Int16 *val2, Int16 *val3, Int8 delim)
{
	int i;
	int start,end;
	int v1,v2,v3;
	int delimcount;


	for (i = *ss; i < *ee; i ++)
	{
		if (!IsDigit(s[i]) && (s[i] != delim))
			return false;
	}

	v1 = v2 = v3 = 0;
	for (start = *ss; start >= 0; start --)
	{
		if (!IsDigit(s[start]) && s[start] != delim)
			break;
	}
	start ++;
	for (end = *ee; s[end]; end ++)
	{
		if (!IsDigit(s[end]) && s[end] != delim)
			break;
	}

	// check if this really looks like a date string
	delimcount = 0;
	for (i = start; i < end; i ++)
	{
		if (s[i] == delim)
			delimcount ++;
		else if (!IsDigit(s[i]))
			return false;
	}
	if (delimcount != 2)
		return false;

	for (i = start; i < end && s[i] != delim; i ++)
	{
		v1 = v1 * 10 + s[i] - chrDigitZero;
	}
	i++;
	for(; i < end && s[i] != delim; i ++)
	{
		v2 = v2 * 10 + s[i] - chrDigitZero;
	}
	i++;
	for (; i < end; i ++)
	{
		v3 = v3 * 10 + s[i] - chrDigitZero;
	}
	*val1 = v1;
	*val2 = v2;
	*val3 = v3;
	
	*ss = start;
	*ee = end;
	return true;
}


/***********************************************************************
 *
 * FUNCTION:    MatchDate
 *
 * DESCRIPTION: This routine match URL pattern.
 *
 * PARAMETERS:  string - pointer to the string
 *				size - string size
 *				pos - start position to match
 *				creator - creator ID of datebook application
 *
 * RETURNED:	true - match URL
 *				false - no URL
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/30/02	Initial Revision (from PalmWiki)
 *
 ***********************************************************************/
static Boolean MatchDate (Char *string, UInt16 size, UInt16 pos,
 UInt16 *startP, UInt16 *endP, DateType *date)
{
	Int16		year, month, day;
	Int16		start, end;
	Boolean		match;


	start = end = pos;

	switch (PrefGetPreference (prefDateFormat))
	{
	case dfMDYWithSlashes:		// 12/31/95
		match = Get3Val(string, &start, &end, &month, &day, &year, '/');
		break;
	case dfDMYWithSlashes:		// 31/12/95
		match = Get3Val(string, &start, &end, &day, &month, &year, '/');
		break;
	case dfDMYWithDots:			// 31.12.95
		match = Get3Val(string, &start, &end, &day, &month, &year, '.');
		break;
	case dfDMYWithDashes:		// 31-12-95
		match = Get3Val(string, &start, &end, &day, &month, &year, '-');
		break;
	case dfYMDWithSlashes:		// 95/12/31
		match = Get3Val(string, &start, &end, &year, &month, &day, '/');
		break;
	case dfYMDWithDots:			// 95.12.31
		match = Get3Val(string, &start, &end, &year, &month, &day, '.');
		break;
	case dfYMDWithDashes:		// 95-12-31
		match = Get3Val(string, &start, &end, &year, &month, &day, '-');
		break;
	case dfMDYWithDashes:		// 12-31-95
		match = Get3Val(string, &start, &end, &month, &day, &year, '-');
		break;
	default:	
		match = Get3Val(string, &start, &end, &year, &month, &day, '/');
		break;
	}

	if (match && month >= 1 && month <= 12 && day >= 1 && day <= 31)
	{
		if (year <= lastYear - 2000)
			year += 2000;
		else if (year < 100)
			year += 1900;

		date->year = year - firstYear;
		date->month = month;
		date->day = day;
		
		*startP = start;
		*endP = end;

		return (true);
	}

	return (false);
}


/***********************************************************************
 *
 * FUNCTION:    CheckInsertionPosition
 *
 * DESCRIPTION: This routine checks  insertion position.
 *
 * PARAMETERS:	fldP - Pointer to the field type object
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static Boolean CheckInsertionPosition (FieldPtr fldP, UInt16 singleTap)
{
	UInt16		selStart, selEnd;
	UInt32		ftrStart, ftrEnd;
	UInt32		ftrValue;


	FldGetSelection (fldP, &selStart, &selEnd);
	if (singleTap)
	{
		if (selStart != selEnd)
			return (false);
	}
	else
	{
		FtrGet (appFileCreator, FeatureSelect, &ftrValue);
		if (ftrValue == IsON)
		{
			FtrGet (appFileCreator, FeatureSelStart, &ftrStart);
			FtrGet (appFileCreator, FeatureSelEnd, &ftrEnd);

			FtrSet (appFileCreator, FeatureSelStart, selStart);
			FtrSet (appFileCreator, FeatureSelEnd, selEnd);

			if (selStart != selEnd || ftrStart > selStart || ftrEnd < selEnd)
				return (false);

			FtrSet (appFileCreator, FeatureSelect, IsOFF);
		}
		else
		{
			FtrSet (appFileCreator, FeatureSelect, IsON);
			FtrSet (appFileCreator, FeatureSelStart, selStart);
			FtrSet (appFileCreator, FeatureSelEnd, selEnd);
			return (false);
		}
	}

	return (true);
}


/***********************************************************************
 *
 * FUNCTION:    LinkApps
 *
 * DESCRIPTION: This routine handles link action.
 *
 * PARAMETERS:	fldP - Pointer to the field type object
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			smb		12/30/03	Initial Revision
 *			smb		02/07/04	Modify to deal with a ToDo application
 *			smb		02/17/04	Modify to fix bug.
 *
 ***********************************************************************/
static void LinkApps (FieldPtr fldP)
{
	PsLinkPreferenceType	prefs;
	Char		buf[MAXLEN_LINKSTRING + 1];
	Char		*string, *matchString;
	UInt16		i;
	UInt16		size;
	UInt16		pos;
	UInt16		matchStart, matchEnd;
	UInt16		len;
	UInt16		launchApp = NONE;
	UInt16		start[NUM_LINK_APPS], end[NUM_LINK_APPS];
	Boolean		launch[NUM_LINK_APPS];
	DateType	date;
	UInt16		tagApp[] = {MEMO, TODO, ADDRESS, MORE1, MORE2, MORE3, MORE4, MORE5};
	UInt16		compAppNum;
	Boolean		jump;


	GetPreferences (&prefs);

	// Check cursor position
	if (!CheckInsertionPosition (fldP, prefs.singleTap))
		return;

	// Get field string
	string = FldGetTextPtr (fldP);
	size = FldGetTextLength (fldP);
	pos = FldGetInsPtPosition (fldP);

	if (string == NULL || pos <= 0 || pos >= size)
		return;

	// Find start and end position of string to be matched.
	for (matchStart = pos; matchStart > 0; matchStart --)
	{
		if (string[matchStart - 1] == chrLineFeed)
			break;
	}
	for (matchEnd = pos; matchEnd < size; matchEnd ++)
	{
		if (string[matchEnd] == chrLineFeed || string[matchEnd] == chrNull)
			break;
	}
	matchString = string + matchStart;
	pos -= matchStart;
	size = matchEnd - matchStart;

	// Matches tag and decides an application to be launched.
	for (i = 0; i < LENGTH(tagApp); i ++)
	{
		if (prefs.creator[tagApp[i]] != LINKAPP_NONE_ID)
		{
			launch[tagApp[i]] = MatchTag (matchString, size, pos,
		 				 prefs.tag[tagApp[i]][START], prefs.tag[tagApp[i]][END],
						 &start[tagApp[i]], &end[tagApp[i]]);
			if (launch[tagApp[i]] && launchApp == NONE)
			{
				launchApp = tagApp[i];
				compAppNum = i;
			}
		}
		else
			launch[tagApp[i]] = false;
	}

	if (launchApp != NONE)
	{
		for (i = compAppNum + 1; i < LENGTH(tagApp); i ++)
		{
			if (launch[tagApp[i]])
			{
				if (start[launchApp] == start[tagApp[i]])
				{
					if (end[launchApp] == end[tagApp[i]])
					{
						if (StrLen (prefs.tag[launchApp][START]) < StrLen (prefs.tag[tagApp[i]][START])
						 || StrLen (prefs.tag[launchApp][END]) < StrLen (prefs.tag[tagApp[i]][END]))
							launchApp = tagApp[i];
					}
					else if (end[launchApp] > end[tagApp[i]])
						launchApp = tagApp[i];
				}
				else if (start[launchApp] < start[tagApp[i]])
					launchApp = tagApp[i];
			}
		}
	}
	else if (prefs.creator[MEMO] != LINKAPP_NONE_ID && prefs.memoWikiName == IsON
	 && MatchWikiName (matchString, pos, &start[MEMO], &end[MEMO]))
		launchApp = MEMO;
	else if (prefs.creator[WEB] != LINKAPP_NONE_ID
	 && MatchURL (matchString, size, pos, TAG_URLSTART, &start[WEB], &end[WEB]))
		launchApp = WEB;
	else if (prefs.creator[MAIL] != LINKAPP_NONE_ID
	 && MatchMailTo (matchString, size, pos, &start[MAIL], &end[MAIL]))
		launchApp = MAIL;
	else if (prefs.creator[DATE] != LINKAPP_NONE_ID
	 && MatchDate (matchString, size, pos, &start[DATE], &end[DATE], &date))
		launchApp = DATE;

	// Launches application if match URL
	if (launchApp != NONE)
	{
		if (launchApp != MAIL)
		{ // Save link stack
			LinkAction (LINKJUMP, string);
		}

		len = end[launchApp] - start[launchApp];
		len = min (len, MAXLEN_LINKSTRING);
		StrNCopy (buf, matchString + start[launchApp], len);
		buf[len] = chrNull;

		switch (launchApp)
		{
		case MEMO:
			jump = LaunchMemo (prefs.creator[MEMO], buf, &prefs);
			break;
		case TODO:
			jump = LaunchToDo (prefs.creator[TODO], buf, &prefs);
			break;
		case ADDRESS:
			jump = LaunchAddress (prefs.creator[ADDRESS], buf);
			break;
		case WEB:
			jump = LaunchWeb (prefs.creator[WEB], buf);
			break;
		case MAIL:
			jump = LaunchMail (prefs.creator[MAIL], buf);
			break;
		case DATE:
			jump = LaunchDate (prefs.creator[DATE], date);
			break;
		default:
			jump = LaunchApp (prefs.creator[launchApp], prefs.dbName[launchApp],
			        buf, prefs.dbLink[launchApp], prefs.dbPrefix[launchApp],
			        prefs.linkChrDBname, prefs.linkChrBookmark);
			break;
		}

		if (!jump)
			LinkAction (LINKFAIL, NULL);
	}
}


/***********************************************************************
 *
 * FUNCTION:    CreatorEventHandle
 *
 * DESCRIPTION: This routine is event handler
 *					 which response to creator notification.
 *
 * PARAMETERS:	notifyParamsP - parameter of notification
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			smb		12/29/03	Initial Revision
 *
 ***********************************************************************/
Err CreatorEventHandle (SysNotifyParamType *notifyParamsP)
{
	UInt16		cardNo;
	LocalID		dbID;
	EventPtr	eventP;
	TablePtr	tableP;
	FieldPtr	fldP = NULL;


	SysCurAppDatabase (&cardNo, &dbID);
	if (!SysNotifyUnregister (cardNo, dbID, appFileCreator, sysNotifyNormalPriority))
	{
  		eventP = (EventPtr)notifyParamsP->notifyDetailsP;
		switch (Swap16 (eventP->eType))
		{
		case fldEnterEvent:
			fldP = (FieldPtr)GetObjectPtr (Swap16(eventP->data.fldEnter.fieldID));
			break;
		case tblSelectEvent:
			if ((tableP = (TablePtr)GetObjectPtr (Swap16(eventP->data.tblSelect.tableID))) != NULL)
				fldP = TblGetCurrentField (tableP);
			break;
		}

		if (fldP != NULL)
			LinkApps (fldP);

		notifyParamsP->handled = true;
	}

	return errNone;
}


/***********************************************************************
 *
 * FUNCTION:    DequeuedEventHandle
 *
 * DESCRIPTION: This routine is event handler
 *					 which response to DequeuedEvent.
 *
 * PARAMETERS:	notifyParamsP - parameter of notification
 *
 * RETURNED:	nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		12/29/03	Initial Revision
 *			smb		02/04/04	Modify to check if selected table item has a field.
 *			smb		06/11/04	Fix a bug; don't work up/down button while PsLink is activated.
 *			smb		12/17/04	Change with modified GetObjectPtr.
 *
 ***********************************************************************/
Err DequeuedEventHandle (SysNotifyParamType *notifyParamsP)
{
	UInt16		cardNo;
	LocalID		dbID;
	EventPtr	eventP;
	EventType	curEvent;
	UInt32		value;
	TablePtr	tableP;
	FieldPtr	fldP;
	KeyDownEventDataType	*backKey, *forwardKey;
	PsLinkPreferenceType	prefs;


	eventP = (EventPtr)notifyParamsP->notifyDetailsP;
	switch (Swap16 (eventP->eType))
	{
	case tblSelectEvent:
		if ((tableP = (TablePtr)GetObjectPtr (Swap16(eventP->data.tblSelect.tableID))) == NULL
		 || (fldP = TblGetCurrentField (tableP)) == NULL)
		{
			notifyParamsP->handled = true;
			break;
		}

	case fldEnterEvent:
		// Register Custom Notification
		SysCurAppDatabase (&cardNo, &dbID);
		SysNotifyRegister (cardNo, dbID,
			appFileCreator, NULL, sysNotifyNormalPriority, NULL);

		// Broadcast Custom Notification
		SysNotifyParamType notify;
		notify.notifyType = appFileCreator;
		notify.broadcaster = appFileCreator;
		notify.notifyDetailsP = notifyParamsP->notifyDetailsP;
		notify.userDataP = NULL;
		notify.handled = false;
		notify.reserved2 = NULL;
		SysNotifyBroadcastDeferred (&notify, sizeof(SysNotifyAppLaunchOrQuitType));

		notifyParamsP->handled = true;
		break;

	case keyDownEvent:
		MemSet (&curEvent, sizeof (EventType), 0);
		curEvent.eType = (eventsEnum)Swap16 (eventP->eType);
		curEvent.data.keyDown.chr = Swap16 (eventP->data.keyDown.chr);
		curEvent.data.keyDown.keyCode = Swap16 (eventP->data.keyDown.keyCode);
		curEvent.data.keyDown.modifiers = Swap16 (eventP->data.keyDown.modifiers);

		if (EvtKeydownIsVirtual (&curEvent))
		{
			if (IsFiveWayNavPalmEvent2(&curEvent))
			/* Was
			if (IsFiveWayNavEvent(&curEvent))
			*/
			{
				if (IsFiveWayNavPressed2(&curEvent))
				{// Set feature value of FeatureNavPressed to "IsON", if 5-way navi pressed.
					FtrSet (appFileCreator, FeatureNavPressed, IsON);
				}

				FtrGet (appFileCreator, FeatureNavPressed, &value);
				if (value != IsON)
				{
					MemSet (eventP, sizeof (EventType), 0);
					return errNone;
				}

				if (NavSelectDirectionPressed(&curEvent, Left))
				{// Select and Left button pressed
					if (JumpApp (LINKBACKWARD))
						goto JUMPED;
					else
						goto NOTJUMPED;
				}
				else if (NavSelectDirectionPressed(&curEvent, Right))
				{// Select and Right button pressed
					if (JumpApp (LINKFORWARD))
						goto JUMPED;
					else
						goto NOTJUMPED;
				}
			}
			else if (curEvent.data.keyDown.chr == vchrJogBack)
			{// JogBack pressed
				if (JumpApp (LINKBACKWARD))
					goto JUMPED;
				else
					goto NOTJUMPED;
			}
		}

		FtrGet (appFileCreator, FeatureBackKey, (UInt32 *)&backKey);
		FtrGet (appFileCreator, FeatureForwardKey, (UInt32 *)&forwardKey);
		if (backKey->chr != 0 && curEvent.data.keyDown.chr == backKey->chr
		 && (backKey->modifiers == 0 || curEvent.data.keyDown.modifiers & backKey->modifiers))
		{
			if (JumpApp (LINKBACKWARD))
				goto JUMPED;
			else
				goto NOTJUMPED;
		}
		else if (forwardKey->chr != 0 && curEvent.data.keyDown.chr == forwardKey->chr
		 && (forwardKey->modifiers == 0 || curEvent.data.keyDown.modifiers & forwardKey->modifiers))
		{
			if (JumpApp (LINKFORWARD))
				goto JUMPED;
			else
				goto NOTJUMPED;
		}
		break;

	default:
		break;
	}

	return errNone;

	NOTJUMPED:
	GetPreferences (&prefs);
	if (prefs.ignoreBtn != IsON)
		return errNone;

	JUMPED:
	FtrSet (appFileCreator, FeatureNavPressed, IsOFF);
	MemSet (eventP, sizeof (EventType), 0);
	notifyParamsP->handled = true;
	return errNone;
}

