/******************************************************************************
 *
 * Copyright (c) 2004 PalmSpace/Sembe.
 *				All rights reserved.
 *
 * File: linkaction.c
 *
 * Description:
 *
 * History:
 *		Jan 16, 2004	Created by sembe
 *
 *****************************************************************************/

#include "PsLink.h"
#include "PsLinkLib.h"
#include "linkaction.h"


/***********************************************************************
 *
 * FUNCTION:    SetLinkStack
 *
 * DESCRIPTION: Save the current state of the linkstack.
 *
 * PARAMETERS:  linkStack
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/29/04	Initial Revision
 *
 ***********************************************************************/
static void SetLinkStack (LinkStackType *linkStack)
{

	PrefSetAppPreferences (appFileCreator, linkStackPrefID, linkStackPrefVersionNum, 
			linkStack, sizeof (LinkStackType), false);
}


/***********************************************************************
 *
 * FUNCTION:	GetLinkStack
 *
 * DESCRIPTION:	Get the current state of the linkstack.
 *
 * PARAMETERS:  linkStack
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		02/29/04	Initial Revision
 *
 ***********************************************************************/
static void GetLinkStack (LinkStackType *linkStack)
{
	UInt16		size;

	size = sizeof (LinkStackType);
	if (PrefGetAppPreferences (appFileCreator, linkStackPrefID, linkStack, &size, false) != linkStackPrefVersionNum)
	{
		MemSet (linkStack, sizeof(LinkStackType), 0);
		SetLinkStack (linkStack);
	}
}


/***********************************************************************
 *
 * FUNCTION:    GetCurrentRecordID
 *
 * DESCRIPTION: This routine gets uniqueID of current record.
 *
 * PARAMETERS:  dbP - pointer to the database
 *				matchString - string to match record
 *				uniqueID - unique ID of current record
 *
 * RETURNED:	true if found
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		01/18/04	Initial Revision
 *			smb		02/07/04	Improve speed
 *			smb		06/12/04	Detach GetBusyRecordID()
 *			smb		03/26/05	Deal with ContactsDB-PAd and CalendarDB-PDa
 *
 ***********************************************************************/
static Boolean GetCurrentRecordID (DmOpenRef dbP, Char *matchString, UInt32 *uniqueIDP)
{
	UInt16		index;
	UInt32		uniqueID;


	*uniqueIDP = NO_UNIQUEID;

	if (dbP == NULL)
		return (false);

	if (GetBusyRecordID (dbP, &uniqueID))
	{
		*uniqueIDP = uniqueID;
		return (true);
	}

	if (matchString != NULL && *matchString != NULL)
	{// Try to match string which should be in the record.
		if (FindRecordByString (dbP, matchString, &index)
		 && DmRecordInfo (dbP, index, NULL, &uniqueID, NULL) == errNone)
		{
			*uniqueIDP = uniqueID;
			return (true);
		}
	}
	else
	{// Save the same unique id as a linked db.
		LinkStackType	linkStack;
		LinkParamsType	*linkedApp;
		DmOpenRef		curDB;
		UInt32			creator;

		GetLinkStack (&linkStack);
		linkedApp = &linkStack.linkedApp;

		if (DmDatabaseInfo (linkedApp->appCardNo, linkedApp->appID, NULL, NULL, 
			NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &creator) == errNone
		 && (creator == 'PAdd' || creator == 'PDat')
		 && CheckDBOpenByCreator (&curDB, 'DATA', creator, false)
		 || CheckDBOpenByID (&curDB, linkedApp->dataCardNo, linkedApp->dataID, false))
		{
			*uniqueIDP = linkedApp->uniqueID;
			return (true);
		}
	}

	return (false);
}


/***********************************************************************
 *
 * FUNCTION:    GetCurrentAppInfo
 *
 * DESCRIPTION: This routine gets current application's information
 *
 * PARAMETERS:  prefs - preferences
 *
 * RETURNED:	true if found
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		01/18/04	Initial Revision
 *			smb		02/23/04	Bug-fix; compatibility with BtToggle. 
 *
 ***********************************************************************/
static Boolean GetCurrentAppInfo (LinkStackType *linkStack, Char *linkString)
{
	Err				err;
	DmOpenRef		db;
	LocalID			dbID;
	UInt16			cardNo;
	UInt32			type;
	UInt32			creator;
	UInt32			uniqueID;
	LinkParamsType	*link;
	UInt16			attr;


	link = &(linkStack->linkstack[linkStack->linkstackindex]);
	link->appCardNo = link->dataCardNo = 0;
	link->appID = link->dataID = NULL;
	link->uniqueID = NO_UNIQUEID;

	// Check all the open database
	db = DmNextOpenDatabase(NULL);
	while (db)
	{
		err = DmOpenDatabaseInfo (db, &dbID, NULL, NULL, &cardNo, NULL);
		if (err == errNone && dbID)
		{
			DmDatabaseInfo (cardNo, dbID, NULL, &attr, NULL, NULL,
					 NULL, NULL, NULL, NULL, NULL, &type, &creator);
			if (type == sysFileTApplication && !(attr & dmHdrAttrResetAfterInstall))
			{// Check attributes to exclude locked resource like "SelectFont" ???
				link->appCardNo = cardNo;
				link->appID = dbID;
			}
			else if (IsData(type) && !(attr & dmHdrAttrResDB))
			{// Check attributes to exclude resource database like "BtToggle icons" ???
				if (GetCurrentRecordID (db, linkString, &uniqueID))
				{
					link->dataCardNo = cardNo;
					link->dataID = dbID;
					link->uniqueID = uniqueID;
				}
			}
		}
		db = DmNextOpenDatabase (db);
	}
	if (link->appID == NULL)
	{// No application found
		return false;
	}

	return true;
}


/***********************************************************************
 *
 * FUNCTION:	SaveLinkedDBInfo
 *
 * DESCRIPTION: This routine saves information of DB to be linked.
 *
 * PARAMETERS:	
 *
 * RETURNED:
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		01/17/04	Initial Revision
 *			smb		02/23/04	Modify to deal with creator
 *
 ***********************************************************************/
void SaveLinkedDBInfo (UInt32 creator, UInt16 appCardNo, LocalID appID,
					 UInt16 dbCardNo, LocalID dbID, UInt16 recordNum)
{
	LinkStackType			linkStack;
	DmSearchStateType		searchState;
	DmOpenRef				dbP;
	UInt32					uniqueID;
	Boolean					open;


	GetLinkStack (&linkStack);

	if (appID == NULL && creator != NULL)
		DmGetNextDatabaseByTypeCreator (true, &searchState,
		 sysFileTApplication, creator, true, &appCardNo, &appID);

	linkStack.linkedApp.appCardNo = appCardNo;
	linkStack.linkedApp.appID = appID;
	linkStack.linkedApp.dataCardNo = dbCardNo;
	linkStack.linkedApp.dataID = dbID;
	linkStack.linkedApp.uniqueID = NO_UNIQUEID;

	if (dbID != NULL)
	{
		open = CheckDBOpenByID (&dbP, dbCardNo, dbID, true);
		if (dbP != NULL)
		{
			if (recordNum != NO_RECORD
			 && !DmRecordInfo (dbP, recordNum, NULL, &uniqueID, NULL))
				linkStack.linkedApp.uniqueID = uniqueID;
			if (!open)
				DmCloseDatabase (dbP);
		}
	}

	SetLinkStack (&linkStack);
}


/***********************************************************************
 *
 * FUNCTION:    MatchCurrentApp
 *
 * DESCRIPTION: This routine matches a current application and the stac.
 *
 * PARAMETERS:	prefs
 *
 * RETURNED:	true if match
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		01/19/04	Initial Revision
 *			smb		06/12/04	Modify to upport MemosDB-PMem
 *
 ***********************************************************************/
static Boolean MatchCurrentApp (LinkStackType *linkStack, Char *linkString)
{
	LinkParamsType	*linkedApp;
	UInt32			uniqueID;
	DmOpenRef		dbP = NULL;
	Boolean			enableJump = false;


	linkedApp = &linkStack->linkedApp;

	if (linkedApp->dataID != NULL)// && linkedApp->uniqueID != NO_UNIQUE_ID)
	{
		if (CheckDBOpenByID (&dbP, linkedApp->appCardNo, linkedApp->appID, false))
		{
			// if the DB which is currently opened is not 'memo',
			// or match unique id of the current record and stacked one.
			if (!CheckDBOpenByCreator (&dbP, 'DATA', sysFileCMemo, false)
			 && !CheckDBOpenByCreator (&dbP, 'DATA', 'PMem', false)
			 || (GetBusyRecordID (dbP, &uniqueID)
			 && uniqueID == linkedApp->uniqueID))
				enableJump = true;
		}
	}
	else
	{
		if (CheckDBOpenByID (&dbP, linkedApp->appCardNo, linkedApp->appID, false))
			enableJump = true;
	}

	return (enableJump);
}


/***********************************************************************
 *
 * FUNCTION:    JumpApp
 *
 * DESCRIPTION: This routine handles to jump other apprication.
 *
 * PARAMETERS:	action
 *
 * RETURNED:	true if succeed to jump
 *
 * REVISION HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		01/16/04	Initial Revision
 *			smb		03/26/05	Deal with ContactsDB-PAd and CalendarDB-PDa
 *
 ***********************************************************************/
Boolean JumpApp (LinkActionType action)
{
	Err						err;
	LinkStackType			linkStack;
	LinkParamsType			*link;
	DmOpenRef				db;
	UInt16					recordNum;
	UInt32					creatorID;
	Boolean					open = false;


	if (!LinkAction (action, NULL))
		return false;

	GetLinkStack (&linkStack);
	link = &(linkStack.linkstack[linkStack.linkstackindex]);
	if (link->appID == NULL)
		return false;

	DmDatabaseInfo (link->appCardNo, link->appID, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL, NULL, &creatorID);

	// if ContactsDB-PAd or CalendarDB-PDa is opened, close it.
	if (creatorID == 'PAdd' && CheckDBOpenByCreator (&db, 'DATA', 'PAdd', false)
	 || creatorID == 'PDat' && CheckDBOpenByCreator (&db, 'DATA', 'PDat', false))
		DmCloseDatabase (db);

	// Check if the database open
	open = CheckDBOpenByID (&db, link->dataCardNo, link->dataID, true);

	// Find record number by uniqueID 
	if (db == NULL || link->uniqueID == NO_UNIQUEID
	 || DmFindRecordByID (db, link->uniqueID, &recordNum) != errNone)
		recordNum = NO_RECORD;

	if (!open && db != NULL)
		DmCloseDatabase (db);

	// Re-open ContactsDB-PAd or CalendarDB-PDa if it was opened.
	if ((creatorID == 'PAdd' || creatorID == 'PDat')
	 && DmOpenDatabaseByTypeCreator ('DATA', creatorID, dmModeReadWrite) == NULL)
	 	return false;

	// Launches application
	if (recordNum == NO_RECORD)
	{// No record was found, just launch a specific application.
		err = SysUIAppSwitch (link->appCardNo, link->appID, sysAppLaunchCmdNormalLaunch, NULL);
	}
	else
	{// Jump to a specific record
		err = SwitchApp (link->appCardNo, link->appID, 0, link->dataCardNo, link->dataID, recordNum, 0, 0, (creatorID == 'NOtl')? 0x0000044c : 0);
	}

	if (err)
		return false;

	// Save db information to be linked
	SaveLinkedDBInfo (NULL, link->appCardNo, link->appID, link->dataCardNo, link->dataID, recordNum);
	return true;
}


/***********************************************************************
 *
 * FUNCTION:    LinkAction
 *
 * DESCRIPTION: This routine backup or restore link history.
 *
 * PARAMETERS:	action
 *
 * RETURNED:	true, if action is available.
 *
 * HISTORY:
 *			Name	Date		Description
 *			----	----		-----------
 *			smb		01/16/04	Initial Revision (from PalmWiki)
 *
 ***********************************************************************/
Boolean LinkAction (LinkActionType action, Char *linkString)
{
	LinkStackType	linkStack;
	Boolean			enableJump;
	Int16			i;


	GetLinkStack (&linkStack);
	enableJump = MatchCurrentApp (&linkStack, linkString);

	switch (action)
	{
	case LINKJUMP:
		if (!enableJump)
			linkStack.linkstackindex = 0;

		if (!GetCurrentAppInfo (&linkStack, linkString))
			return false;
		if (linkStack.linkstackindex + 1 < MAXLINK)
		{
			linkStack.linkstackindex ++;
		}
		else
		{
			for (i = 0; i < linkStack.linkstackindex; i ++)
			{
				linkStack.linkstack[i] = linkStack.linkstack[i + 1];
			}
		}
		linkStack.linkstackmax = linkStack.linkstackindex;
		break;

	case LINKFAIL:
		if (linkStack.linkstackindex > 0)
			linkStack.linkstackindex --;
		linkStack.linkstackmax = linkStack.linkstackindex;
		break;

	case LINKFORWARD:
		if (!enableJump)
			linkStack.linkstackindex = linkStack.linkstackmax;

		if (linkStack.linkstackindex >= linkStack.linkstackmax ||
		   linkStack.linkstackindex + 1 >= MAXLINK)
			return false;
		linkStack.linkstackindex++;
		break;

	case LINKBACKWARD:
		if (!enableJump)
			linkStack.linkstackindex = 0;

		if (linkStack.linkstackindex <= 0)
			return false;
		if (linkStack.linkstackindex == linkStack.linkstackmax)
		{
			if (!GetCurrentAppInfo (&linkStack, linkString))
				return false;
		}
		linkStack.linkstackindex --;
		break;
	}

	SetLinkStack (&linkStack);
	return true;
}

